

from .config     import Config


# ==================================================================================================
# Ranking
# ==================================================================================================

def rank_without_group(df, col_score, find='max', col_target='', on_conflict='first'):
    '''
    Select the max row. Or the min.  
    Or mark the rows instead of selecting them. 
    * col_score:    Name of the column whose minimum or maximum is to be found.
    * find:         'min' or 'max', default is 'max'.
    * col_target:   Should a ranking column be added? If yes, then give the name here.
                    If col_target is given, then the dataframe is returned completely, plus the new column.
                    If no col_target is given, only one row with rank 1 is returned (on_conflict is ignored).  
    '''
    
    # Nur Extremzeile ausgeben    
    if col_target=='':
        if find=='max':
            idx = df[col_score].idxmax()
        else:
            idx = df[col_score].idxmin()  
        
        if idx is None:
            return None
        return df.loc[[idx]]

      
    # rank Spalte erzeugen und alle Zeilen ausgeben
    else:
        if find=='max':
            ascending = False
        else:
            ascending = True          
        result = df.copy()
        result[col_target] = result[col_score].rank(method=on_conflict, ascending=ascending)
        maxrank = int('9'*len(str(df.shape[0]+1)))
        result[col_target] = result[col_target].fillna(maxrank)
        if on_conflict != 'average':
            result[col_target] = result[col_target].astype(int)    
        return result
    

def rank(df, col_score, cols_group=None, find='max', col_target='', on_conflict='first',verbose=None):
    ''' 
    Select the max row per group. Or the min.
    Or mark the rows instead of selecting them. 
    * cols_group:   Name of the columns to be grouped by. None if no grouping needed.
    * col_score:    Name of the column whose minimum or maximum is to be found.
    * find:         'min' or 'max', default is 'max'.
    * col_target:   Should a ranking column be added? If yes, then give the name here.
                    If col_target is given, then the dataframe is returned completely, plus the new column.
                    If no col_target is given, only rank 1 is returned.
    * on_conflict:  How to rank the group of records that have the same value. 
                    Possible values: 'min','max','average','dense' and 'first', see pandas rank.
    '''    
    
    if verbose is None:
        verbose = Config.get('VERBOSE')  
        
    if cols_group is None:
        result = rank_without_group(df=df, col_score=col_score, find=find, col_target=col_target, on_conflict=on_conflict)
        if verbose:
            n0 = df.shape[0]
            n1 = result.shape[0]        
            print( 'rank: {0} rows less, now {1} rows'.format(n0-n1, n1)  )
        return result
    
    if find=='max':
        ascending = False
    else:
        ascending = True    
            
    if col_target=='':
            result = df.sort_values(by=col_score, ascending=ascending, kind='mergesort').drop_duplicates(cols_group)
            
    else:
        result = df.copy()
        result[col_target] = result.groupby(cols_group)[col_score].rank(method=on_conflict, ascending=ascending)
        maxrank = int('9'*len(str(df.shape[0]+1)))        
        result[col_target] = result[col_target].fillna(maxrank)
        if on_conflict != 'average':
            result[col_target] = result[col_target].astype(int)         
        
    if verbose:
        n0 = df.shape[0]
        n1 = result.shape[0]        
        print( 'rank: {0} rows less, now {1} rows'.format(n0-n1, n1)  )        
    return result 

    
    
# Alternative Methode für rank   
# see # https://stackoverflow.com/questions/50381064/select-the-max-row-per-group-pandas-performance-issue  
# ok wenn wenige große Gruppen
# ok wenn viele kleine Gruppen
# Ergebnis ist sortiert    
#
def rank_sort(df, cols_group, col_score, find='max'):
    
    if find=='max':
        keep = 'last'
    else:
        keep = 'first'
        
    sort_spalten = cols_group + [col_score]
    found        = df.sort_values(sort_spalten).drop_duplicates(cols_group, keep=keep).index
    
    return df.loc[found]    
    
    
    
# Alternative Methode für rank       
# rank_idxminmax:
# sehr schnell wenn wenige große Gruppen
# extrem langsam wenn viele kleine Gruppen
# Ergebnis ist sortiert    
def rank_idxminmax(df, cols_group, col_score, find='max'):
    
    if find=='max':
        found = df.groupby(cols_group)[col_score].idxmax()
    else:
        found = df.groupby(cols_group)[col_score].idxmin()
            
    return df.loc[found]       





    
import random    
import numpy  as np
from .config   import Config
from .pandas  import repeat


def gmean(x):
    '''
    Calculates Geometric Mean.
    NaN-safe.
    Usable for values with logarithmic character, e.g. the Geometric Mean of [1, 10, 100] is 10. 
    '''
    return np.exp(np.mean(np.log(x)))
    
    
    
# ==================================================================================================
# scale
# ==================================================================================================
     
    
def scale(series, method, powerfactor=1, almostzero=0.00000001, verbose=None ):
    '''
    Scales all values of a numeric series to a defined value range.
    * method must be 'max_abs','min_max','min_max_robust',
      'rel','mean','median','compare_median','rank' or 'random'
    * powerfactor is an additional parameter for scaling 'rank'

    ## scale method='max_abs'
    * scales every value with a fixed factor
    * one finds: Every scaled value is somewhere in the range -1..1
    * series_scaled.max() is often 1
    * series_scaled.min() can be anything -1..0.999    

    ## scale method='min_max'
    * forces all scaled values to fit the full range 0..1 (closed interval)
    * one finds: series_scaled.min() == 0
    * one finds: series_scaled.max() == 1 if there are more than 2 different values

    ## scale method='min_max_robust'
    * like min_max, but robust against outlier values. 
      Works with quantile(0.01) and quantile(0.99) instead of min() and max().
      The scaling is therefore not determined by the outliers.
    * scales 98% of the values to fit the range 0..1
    * one finds: series_scaled.min() <= 0 but not much lower  than -0.02 normally
    * one finds: series_scaled.max() >= 1 but not much higher than  1.02 normally

    ## scale method='rel'
    * scales every value with a fixed factor so that
    * series_scaled.sum() == 1  or  series_scaled.sum() == -1 
    * therefore, this scaling is well suited for frequencies, 
      the numerical values are then the relative frequencies
    * series_scaled.min()  and  series_scaled.max() are numbers near 0

    ## scale method='mean'
    * also called z-score
    * one finds: series_scaled.mean() == 0 
    * one finds: series_scaled.std()  == 1 if there are more than 2 different values
    * a typical range for the scaled values is -1.7..1.7

    ## scale method='median'
    * like mean, but more robust against outliner values.
    * one finds: series_scaled.median() == 0 
    * a typical range for the scaled values is -1..1, 
      but the range can be much wider than with mean

    ## scale method='compare_median'
    * scales 0..1 below median and 1.. above median 
    * So you can multipy the scaled values by any factor, keeping the comparison to the median.
    * one finds: all values < median are 0..1
    * one finds: all values == median are 1
    * one finds: all values > median are > 1
    * series_scaled.median() will be very near 1.
    * unlike most other scalings, the plots do not lie on top of each other, 
      but have a different shape

    ## scale method='rank'
    * scales 0..1 (open interval)
    * scales by rank
    * distributes evenly over the interval, the original shape is destroyed. 
      The plots do not lie on top of each other.
    * if powerfactor == 1 one finds: series_scaled.median() == 0.5 
    * additional parameter powerfactor deforms the scale, see example

    ## scale method='random'
    * scales randomly
    * generates an ugly, krank scaling for testings
    
    '''    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
        
    if np.isnan( series.min() ):
        if verbose:
            print('scale: series.min() is NaN')
        return series
    
    if np.isnan( series.max() ):
        if verbose:
            print('scale: series.max() is NaN')
        return series    
    
    # nnan merken
    series_nnan = series.isnull().sum()  
    if verbose and series_nnan > 0:
        print('scale: series_nnan == ', series_nnan)    
        
    if method == 'rel': 
        summe = series.sum() 
        if summe > almostzero:
            return series / summe
        elif summe < -almostzero:
            return -series / summe
        else:
            return  repeat( [0], series.shape[0] )  # only zeros
        
    elif method == 'max_abs':
        nenner = series.abs().max()
        if nenner > almostzero:
            return series  / nenner
        else:
            return  repeat( [0], series.shape[0] )  # only zeros
        
    elif method == 'min_max':
        nenner = (series.max() - series.min())
        if nenner > almostzero:
            return (series - series.min())  /  nenner
        else: 
            return  repeat( [0], series.shape[0] )  # only zeros
    
    
    elif method == 'min_max_robust':
        nenner = (series.quantile(0.99) - series.quantile(0.01)) 
        if nenner > almostzero:
            return (series - series.quantile(0.01))  /  nenner
        else:
            return  repeat( [0], series.shape[0] )  # only zeros
    
    # --------------------------------------------------------
    elif method == 'mean':
        nenner = series.std() 
        if nenner > almostzero:
            result = (series - series.mean()) / nenner
        else:
            return  repeat( [0], series.shape[0] )  # only zeros
        
        if abs(result.mean()) <=  almostzero:
            return result # alles ok
        if almostzero > 0.01  or  result.max() - result.min() < almostzero:
            #print('sinnlos', almostzero)            
            return result
        else:
            #print('next try', almostzero)
            return scale(result, method, almostzero=almostzero*10) 
        
    elif method == 'median': # ist im median 0
        nenner = (series.quantile(0.75) - series.quantile(0.25)) 
        if nenner > almostzero:
            return (series - series.median())  / nenner 
        else:
            return  repeat( [0], series.shape[0] )  # only zeros
        
    elif method == 'rank':
        if powerfactor == 1:
            rang = series.rank(method='dense')
        else:
            rang = np.power(series.rank(method='dense'), powerfactor)
        maximum = rang.max()
        result = rang / maximum
        abziehen = result.min() / 2
        return result - abziehen     
    
    elif method == 'compare_median':
        median = series.quantile(0.5)        
        result = series.copy()
        result = 1 + scale( series, method='median'  )  # erst mal für alle, ist im median 1          
        mask = (series < median)           
        result.loc[mask] = scale( series[mask], method='min_max' )   # kleiner median: 0..1 
        return result
    
    elif method == 'random':
        result = series.copy()
        
        # add
        if result.abs().min() < almostzero  or  result.abs().max() < almostzero: 
            result = result + random.uniform(-10, 10)          
            
        # mult
        würfel = random.randint(1, 7)
        if würfel == 1:
            result = result * random.uniform(-1, 1)           
        elif würfel == 2:
            result = result * random.uniform(-10, 10)   
        elif würfel == 3:
            result = result * random.uniform(-100, 100)    
        elif würfel == 4:
            result = result * random.uniform(-1000, 1000) 
        elif würfel == 5:
            result = result * random.uniform(-0.1, 0.1)   
        elif würfel == 6:
            result = result * random.uniform(-0.01, 0.01) 
        elif würfel == 7:
            result = result * random.uniform(-0.001, 0.001)  
            
        # spread
        diff = result.max() - result.min()
        p = 1       
        if diff < 0.001  and  series_nnan==0:
            p = random.uniform( 1, 4) 
        elif diff < 0.01  and  series_nnan==0:
            p = random.uniform( 1, 2)         
        elif diff < 0.1  and  series_nnan==0:
            p = random.uniform( 1, 1.5)               
        if p != 1:
            result_try = result ** p    
            if series_nnan == result_try.isnull().sum()  :
                result = result_try 
        
        # add again
        if result.abs().min() < almostzero  or  result.abs().max() < almostzero: 
            result = result + random.uniform(-0.1, 0.1)             
                      
        return result











import warnings

import pandas     as pd 
import numpy      as np

   
try:
    import dtale
except ImportError:
    warnings.warn('dtale not importable. Think about pip install dtale')    
    
try:
    from termcolor import colored    # um Fehlermeldungen rot auszugeben
except ImportError:
    pass

from .config   import Config
from .analyse  import col_names    
    
# ==============================================================================================
# search_str
# ==============================================================================================    
    
def search_str(df, find, without=[]):
    """ 
    Searches all str columns of a dataframe.
    Useful for development and debugging.
    * find:    What is to be found?                  String or list of strings.
    * without: Which columns should not be searched? String or list of strings.    
    """
    # Argumente formatieren
    if type(find) is str:
        find = [find] 
    if type(without) is str:
        without = [without]
        
    cols = col_names(df, only='str')
    cols = [c for c in cols if not c in without]
    mask = df[cols].isin(find).any(axis=1)  
    return df[mask]    
        
    
    
# ==============================================================================================
# check_mask
# ==============================================================================================

def check_mask(df, mask, expectation_min=None, expectation_max=None, msg='', stop=True, verbose=None):
    """ 
    Count rows filtered by a binary mask.
    Raises an error, if the number is unexpected.

    Examples:
    ==========
    check_mask( df, mask )         # just show the number of rows   
    check_mask( df, mask, 2000 )   # checks for about 2000 rows (if not 0: not more than double, not less than half)
    check_mask( df, mask, 0 )      # checks for exactly 0 rows    
    check_mask( df, mask, 10, 50)  # checks for 10..50 rows

    Example with later raise:
    =========================
    error = check_mask(df, mask, 214, stop=False)        
    grid(df, mask, error)        
    bpy.raise_if(error)
    """
    
    def print_red(msg):
        try:
            msg = colored(msg, 'red', attrs=['reverse','bold'])
        except:
            pass   
        print(msg)      
    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
    
    error = ''
    
    if msg:
        msg = '{:<50s}'.format(msg)
    
    # expectation_max fehlt
    if not (expectation_min is None) and (expectation_max is None):
        if expectation_min == 0:
            e_min = 0
            e_max = 0
        else:
            e_min = int(expectation_min * 0.5) # Verdoppelung oder Halbierung wird toleriert
            e_max = int(expectation_min * 2.0) + 1        
        return check_mask(df, mask, e_min, e_max, msg=msg, stop=stop, verbose=verbose) 
    
    if type(mask) == pd.Series   or   type(mask) == np.ndarray:
        anz_ds = df[mask].shape[0]
    else:
        anz_ds = df.shape[0]
        
    # expectation_min fehlt
    if expectation_min is None:     
        print('check_mask:', msg, "{0} rows".format(anz_ds).strip()  )   
        return
    
    # wurde vielleicht schon abgearbeitet?
    #if (anz_ds == 0)   and   (expectation_min > 0) and verbose:
    #    print_red(msg + " WARNING: {0} rows, but it should be at least {1}".format(anz_ds, expectation_min)  ) 
    #    return 
    
    if anz_ds > expectation_max:
        error = " ERROR: {0} rows, but it should be a maximum of {1}".format(anz_ds, expectation_max)  
    elif (anz_ds < expectation_min):
        error = " ERROR: {0} rows, but it should be at least {1}".format(anz_ds, expectation_min)     
    else:
        if verbose:
            print('check_mask:',msg, "{0} rows".format(anz_ds).strip()  )  
        
    #raise_later  
    if (anz_ds > expectation_max)  or (anz_ds < expectation_min):
        if stop:
            raise Exception(  (msg + error).strip()  )
            
        else:
            print_red(   (msg + error).strip()  )
            return error.strip()
    
    


# ==============================================================================================
# dtale Settings
# ==============================================================================================

try:
    import dtale.global_state as global_state
except:
    pass


def set_grid(**kwargs):
    ''' 
    Sets settings for dtale.
    Ex:
    set_grid(max_column_width=1000)
    '''
    #print(kwargs)
    global_state.set_app_settings( kwargs) 
    
    
    
def reset_grid():
    '''
    Resets settings for dtale to default values.
    '''
    set_grid( max_column_width=300, 
                  theme='light')
    
    
try:    
    reset_grid()   
except:
    pass
    


# ==============================================================================================
# grid
# ==============================================================================================

def grid( df, mask=None, error='€€€', color='blue', **kwargs ):
    '''
    Visualization of a DataFrame using dtale.
    * df:     DataFrame to show
    * mask:   Binary mask, function or Searchstring to reduce the number of rows
    * error:  Error message
    * color:  Color of the error message.
    * kwargs: Options for dtale. See https://github.com/man-group/dtale#instance-settings
              Caution: This will affect all grids in this notebook...
    dtale may not work in a multiscreen setting on windows.
    
    === Examples ===
    grid(df)                        # show all rows
    grid(df,mask)                   # show with binary mask    
    grid(df,sample)                 # show with functionally mask

    === For error indication after check_mask ===
    error = check_mask(df, mask, 900, stop=False)
    grid(df, mask, error) 
    and later: raise_if(error)
    '''
    
    def print_color(msg, color):
        if color:
            try:
                msg = colored(msg, color, attrs=['reverse','bold'])
            except:
                pass   
        print(msg)  
    
    
    if df is None:
        print_color('Nothing to show', 'red')        
        return 
    
    if df.shape[0] == 0:
        print_color('No rows', color)   
        return
    
    # verschiedene mask behandeln
    if isinstance(mask, pd.Series)  or  isinstance(mask, np.ndarray):
        df_show = df[mask]
    elif callable(mask):   
        df_show = mask(df)
    elif isinstance(mask, str):
        df_show = search_str(df, mask)
    else:
        df_show = df
        
    if df_show.shape[0] == 0:
        print_color('No rows, mask filters them all away', color)
        return


    # widget erzeugen
    #widget = dtale.show(df_show,**kwargs) # funktioniert nicht  
    #widget.update_settings({'theme':'light', 'max_column_width':150}) # funktioniert auch nicht
    #global_state.set_app_settings( kwargs)  
    
    #set_grid(**kwargs)
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)    
        widget = dtale.show(df_show,**kwargs) # funktioniert nicht für alle options. Z.B. nicht für max_column_width
    #reset_grid()
    #widget = dtale.show(df_show)   

    # Anzahl printen
    if error: # gemeint ist, wenn error irgendwas enthält, z.B. den Standardwert'€€€'
        if (df_show.shape[0] != df.shape[0])  and  (df_show.shape[0] > 0):
            print( df_show.shape[0], 'rows out of', df.shape[0] )     
        else:
            print( df_show.shape[0], 'rows' )
                
    # grid wird zur Anzeige potentieller Fehler genutzt, aber es gibt keine Fehler.
    # Dann wird auch kein Widget ausgegeben.
    else:       
        return 
        
    # Widget ausgeben
    if df_show.shape[0] > 0:
        return widget



    


import warnings
import pandas as pd

from functools     import partial  

from .pandas       import dataframe, reset_index, drop_cols, rename_col, move_cols, first_valid_value, last_valid_value
from .aggregate    import most_freq_elt
from .rank         import rank_without_group
from .analyse      import col_names




# ==================================================================================================
# specials
# ==================================================================================================


def specials(df, find=['head','first','min','most','max','nan','last','tail'], indicator=None, sort='index' ):
    '''
    Returns rows representing all special values per column.
    The resulting dataframe has the same minimums and maximums per column as the input dataframe, 
    and if a column in the input contains NaNs the result will contain NaNs as well.
    * find: List of what values are searched. 
      Possible values: 'head','first','min','most','max','nan','last','tail'
    * indicator: Show additional column with a note, why the row is in the result
    '''
    
    if type(find) is str:
        find = [find] #let the command take a string or list 
        
    result = df.head(0)
    result['__indicator__'] = ''
    result['__score__']     = 0
    group_cols = []
   
    # head
    if 'head' in find:   
        try: 
            zeile = df.head(1).copy()
            zeile['__indicator__'] = 'head'
            zeile['__score__']     = 9999            
            result = pd.concat([result,zeile]) 
        except:
            pass   


    for col in col_names(df,sort=True):
        df_col = df[col].dropna()
        
        # first
        if 'first' in find:   
            try:        
                value = first_valid_value(df_col)
                mask = result[col] == value       # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_first'
                else:
                    mask = df[col] == value
                    zeile = df[mask].head(1)
                    zeile['__indicator__'] = col + '_first'
                    result = pd.concat([result,zeile]) 
                if not col in group_cols:
                    group_cols += [col]
            except:
                pass    
            
         

        # min
        if 'min' in find:
            try:
                mask = result[col] == df_col.min() # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_min'
                else:
                    zeile = rank_without_group(df, col, find='min')
                    zeile['__indicator__'] = col + '_min'
                    result = pd.concat([result,zeile])
                if not col in group_cols:
                    group_cols += [col]                    
            except:
                pass
            
        # most
        if 'most' in find:   
            try:        
                value = most_freq_elt(df_col)
                mask = result[col] == value # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_most'
                else:
                    mask = df[col] == value
                    zeile = df[mask].head(1)
                    zeile['__indicator__'] = col + '_most'
                    result = pd.concat([result,zeile]) 
                if not col in group_cols:
                    group_cols += [col]                    
            except:
                pass               

        # max
        if 'max' in find:     
            try:            
                mask = result[col] == df_col.max() # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_max'
                else:
                    zeile = rank_without_group(df, col, find='max')
                    zeile['__indicator__'] = col + '_max'
                    result = pd.concat([result,zeile])  
                if not col in group_cols:
                    group_cols += [col]                    
            except:
                pass                    

        # nan
        if 'nan' in find:   
            try:            
                mask = result[col].isna() # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_nan'
                else:
                    mask = df[col].isna()
                    zeile = df[mask].head(1)
                    zeile['__indicator__'] = col + '_nan'
                    result = pd.concat([result,zeile])  
                if not col in group_cols:
                    group_cols += [col]                    
            except:
                pass     
            

            
        # last
        if 'last' in find:   
            try:        
                value = last_valid_value(df_col)
                mask = result[col] == value # schon drin?
                if result[mask].shape[0] > 0:
                    result.loc[mask,'__indicator__'] += ' ' + col + '_last'
                else:
                    mask = df[col] == value
                    zeile = df[mask].head(1)
                    zeile['__indicator__'] = col + '_last'
                    with warnings.catch_warnings():
                        warnings.simplefilter(action='ignore', category=FutureWarning) 
                        result = pd.concat([result,zeile]) 
                if not col in group_cols:
                    group_cols += [col]                    
            except:
                pass                 
            
    # tail
    if 'tail' in find:   
        try: 
            with warnings.catch_warnings():
                warnings.simplefilter(action='ignore', category=FutureWarning) 
                zeile = df.tail(1).copy()
                zeile['__indicator__'] = 'tail' 
                #zeile['__score__']     = 9998                
                result = pd.concat([result,zeile]) 
        except:
            pass               
                
    # sortieren
    mask = result['__score__'].isna()
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)    
        result.loc[mask,'__score__'] = result[mask]['__indicator__'].str.split().str.len()
    result = result.sort_values('__score__', ascending=False)
    
    # Dups entfernen
    mask = ~result.index.duplicated(keep='first')
    result = result[mask]
    
    #print(group_cols)
    
    if indicator is None:
        result = drop_cols(result,'__indicator__')
    else:
        result = rename_col(result,'__indicator__',indicator)
    
    result = drop_cols(result,'__score__')
    if sort == 'index':
        return result.sort_index()
    else:
        return result




# ==================================================================================================
# sample
# ==================================================================================================


def sample(df, size=None):
    ''' 
    Returns some sample rows.
    Always the beginning and the end, 
    plus the other specials() --see there--,
    plus some random rows.
    * size: Number of rows to return. If size=None, all specials() are returned.
    '''
    if df is None:
        return None
    if size is None:
        return specials(df)
    if size <= 0:
        return df.head(0)
    if df.shape[0] <= size:
        return df
    
    sp = specials(df, sort='score')
    
    # specials sind groß genug
    if sp.shape[0] >= size: 
        return sp.head(size).sort_index()
    
    sa = df.iloc[1:-1].sample(size-2)
    return pd.concat([sp,sa]).head(size).sort_index()


sample_10     = partial(sample, size=10)    
sample_20     = partial(sample, size=20)   
sample_100    = partial(sample, size=100) 
sample_1000   = partial(sample, size=1000) 
sample_10000  = partial(sample, size=10000) 
sample_100000 = partial(sample, size=100000) 



#def sample_notnull(df, size=6):
#    ''' 
#    Returns some sample rows. Prefers notnull rows if possible.
#    Always the beginning and the end, plus some random rows in the middle.
#    * size: Number of rows returned
#    '''    
#    if size <= 0:
#        return df.head(0)
#    if df.shape[0] <= size:
#        return df    
#    df1 = df.sample(size*10, replace=True).dropna()
#    df2 = df.sample(size*10, replace=True).dropna(thresh=2)
#    df3 = df.sample(size,    replace=True)
#    result = pd.concat( [df1, df2, df3] )  
#    result = result[~result.index.duplicated(keep='first')]
#    result = result.head(size).sort_index()
#    return result



       





import numpy  as np
import pandas as pd

from pandas.api.types import is_string_dtype, is_numeric_dtype

from .config       import Config



 

    
#############################################################################################
# ...........................................................................................
# Class values_info
# ...........................................................................................
#############################################################################################
  
# this one returns a similar result:
# df.agg([ 'count', 'nunique', 'min', 'max', nnan ]).transpose()    
#
class values_info:
    """    
    Provides information on the values in a series. 
    Ex: i = values_info(series)     
          i.info() # returns all attributes

    * category_maxsize: How big can a category get to be suggested as datatype_suggest?
    * nanless_ints: Are numpy's integer classes (that don't know NaN) suggested as datatype_suggest?     
    """
    
    def __init__(self, data, category_maxsize=-1, nanless_ints=False):
        
        from pandasklar.analyse   import ntypes
        from pandasklar.type_info import type_info
        
        self.n         = int(data.shape[0])
        self.nnan      = int(data.isna().sum())
        self.ntypes    = ntypes(data)
        try:
            self.nunique   = int(data.nunique())
        except:
            self.nunique   = int(data.apply(str).nunique())
            if self.nnan > 0:
                self.nunique -= 1
        self.ndups     = self.n - self.nunique - self.nnan
        
        if is_numeric_dtype(data):
            self.vmin    = round(float(data.min()),2)
            self.vmax    = round(float(data.max()),2)
            self.vmean   = round(float(data.mean()),2)
            self.vsum    = round(float(data.sum()),2)
            try:
                self.vmedian = round(float(data.quantile(0.5)),2) 
            except:
                self.vmedian = np.NaN
                
        else: 
            data_dropna = data.dropna()
            try:
                self.vmin    = data_dropna.min()
                self.vmax    = data_dropna.max()
            except:
                self.vmin    = np.NaN
                self.vmax    = np.NaN                
            self.vmean   = np.NaN   
            self.vmedian = np.NaN
            self.vsum    = np.NaN
            
        # datatype_suggest
        self.datatype_suggest = ''
        if (self.ntypes == 1):
            self.type_info = type_info(data)  
            
            # int_versuche: Dtypes für Integer
            if self.nnan == 0  and nanless_ints:
                int_versuche = ['np.int8','np.uint8','np.int16','np.uint16','np.int32']
            else:
                int_versuche = ['pd.Int8','pd.Int16','pd.Int32']             
            
            # category oder string?
            if   (self.type_info.name_instance in ['str','_ElementUnicodeResult'])  and  self.type_info.name_short in ['object','string']:
                if (self.nunique < category_maxsize):
                    self.datatype_suggest = 'pd.category'  
                else:
                    self.datatype_suggest = 'pd.string'         
                    
            # float in int?        
            if self.type_info.name_short.startswith('float')   and   self.datatype_suggest=='': #and   self.nnan == 0:    
                try:
                    if (int(self.vmin)-self.vmin == 0)   and    (int(self.vmax)-self.vmax == 0)   and   ((data.astype('Int64')-data).sum() == 0): 
                        for versuch in int_versuche :
                            i = type_info(versuch)            
                            if i.xmin <= self.vmin   and   i.xmax >= self.vmax:
                                self.datatype_suggest = versuch
                                break                       
                except:
                    pass
                
            # float verkleinern?
            if self.type_info.name_short.startswith('float')   and  self.datatype_suggest=='':
                for versuch in ['np.float32'] : # float16 nehmen wir nicht, da kann man keinen Mittelwert berechnen
                    i = type_info(versuch)            
                    if i.xmin < self.vmin   and   i.xmax > self.vmax:
                        self.datatype_suggest = versuch
                        break
            
            # int verkleinern?
            # object in int?
            if (self.type_info.name_short.startswith(('int','uint','Int'))   and   self.datatype_suggest=='')   or   \
               ((self.type_info.name_instance =='int')   and   (self.type_info.name_short == 'object')   and   self.datatype_suggest==''):
                for versuch in int_versuche :
                    i = type_info(versuch)            
                    if i.xmin <= self.vmin   and   i.xmax >= self.vmax:
                        self.datatype_suggest = versuch
                        break   
                                            
            # wieder löschen wenn schon erfüllt
            if self.datatype_suggest == self.type_info.name:
                self.datatype_suggest = ''
        # Ende if (self.ntypes == 1)
        
        
    def info(self):
        """Returns all attributes"""
        result = dict(self.__dict__) # Kopie ziehen
        #del result['data']
        return result    
    
    
    



import collections, warnings

import numpy  as np
import pandas as pd

from .config    import Config
from .pandas    import first_valid_value, last_valid_value




#################################################################################
# ...............................................................................
# Class type_info
# ...............................................................................
#################################################################################
    
class type_info:
    """               
    Provides information about pandas types and standardises them.
    Is initialised with anything, e.g. with the name of a class, or with the class itself.
    Or, even better, with a series.
    Ex:   i = type_info('Int32')     
          i.info()            # returns all attributes, including for example:
          i.class_object      # the class object
          i.name              # the name of the Dtype
          i.name_instance     # type of the contents of the series
          i.instance1         # an example instance that is not NaN              
    """
    
    def __init__(self, search):
        
        #from pandasklar.analyse  import val_first_valid, val_last_valid
        
        # Beispielinstanzen
        self.instance1 = None
        self.instance2 = None
        if isinstance(search, pd.Series): # Es wurde eine Series übergeben
            self.instance1 = first_valid_value(search)
            self.instance2 = last_valid_value(search)     
            search = str(search.dtype)
        elif not isinstance(search, str): # Es wurde eine Klasse übergeben
            search = str(search)
            
        # search ist jetzt str
        self.search        = search
        self.name          = None          
        self.framework     = ''        
        self.name_short    = search
        self.name_long     = None    
        self.class_object  = None
        self.is_hashable   = None
        self.nan_allowed   = True
        self.name_instance = ''
        self.xmin          = None
        self.xmax          = None        

        # framework
        if 'Dtype' in search:
            self.framework = 'pd'  
            
        # name_instance
        if self.instance1 is not None:
            if (str(type(self.instance1)) == str(type(self.instance2))): 
                self.name_instance = type(self.instance1).__name__
            else: 
                self.name_instance = 'mix'
                
        # is_hashable
        if self.name_instance == 'mix':
            self.is_hashable = False    
        else:
            self.is_hashable = isinstance(self.instance1, collections.Hashable)  and  isinstance(self.instance2, collections.Hashable)
            
        # name_short
        if '.' in search:
            self.name_short = search.split('.')[-1]
        self.name_short = self.name_short.replace("<class '","").replace("'>","").replace("numpy.","").replace("Dtype","")       

            
        # name_long
        if self.name_short.startswith( ('int','uint') ):
            self.name_long   = 'np.' + self.name_short
            self.nan_allowed = False 
            self.framework   = 'np'             
            
        elif self.name_short.startswith( 'float' ):
            self.name_long  = 'np.' + self.name_short    
            self.framework  = 'np'     
            
        elif self.name_short.startswith( 'Float' ):
            self.name_long = 'pd.' + self.name_short + 'Dtype'  
            self.framework = 'pd'                 
            
        elif self.name_short.startswith( 'Int' ):
            self.name_long = 'pd.' + self.name_short + 'Dtype'  
            self.framework = 'pd'              
            
        elif self.name_short.startswith( ('str','Str') ):
            self.name_long  = 'pd.StringDtype'  
            self.name_short = 'string' # Namen standardisieren, damit astype ihn versteht
            self.framework  = 'pd'    
            
        elif self.name_short.startswith( 'obj' ):
            self.name_long  = 'object'  
            self.framework  = ''      
            
        elif self.name_short.startswith( ('category','Categorical') ):
            self.name_short = 'category'              
            self.name       = 'pd.category'              
            self.name_long  = 'pd.Categorical'  
            self.framework  = 'pd'                
     
        # name
        if self.framework:
            self.name = self.framework + '.' + self.name_short
        else:
            self.name = self.name_short
            
        # class_object    
        if self.name_long:
            self.class_object = eval(self.name_long)
            
        # xmin und xmax
        #print(self.name_short)
        if self.framework == 'np'  and  self.name_short.startswith( ('int','uint') ):
            self.xmin = np.iinfo(self.class_object).min
            self.xmax = np.iinfo(self.class_object).max   
        elif self.framework == 'pd'  and  self.name_short.startswith( 'Int' ):
            co = eval(  'np.' + self.name_short.lower()  ) # class_object faken
            self.xmin = np.iinfo(co).min  
            self.xmax = np.iinfo(co).max    
        elif self.framework == 'np'  and  self.name_short.startswith( 'float' ):
            self.xmin = np.finfo(self.class_object).min
            self.xmax = np.finfo(self.class_object).max                
            

        
    def info(self):
        """Returns all attributes"""
        result = dict(self.__dict__) # Kopie ziehen
        del result['search']
        return result
    
    





import warnings, copy

import pandas as pd 
import numpy  as np
import bpyth  as bpy
   
from functools   import partial    
from collections import Counter, defaultdict 

from .config       import Config
from .rank         import rank


#from .python     import flatten, rtype, shape, superstrip
#import locale
#locale.setlocale(locale.LC_ALL, '') 



#################################################################################################
# ...............................................................................................
# Einzelner Dataframe
# ...............................................................................................
#################################################################################################




    
# ==================================================================================================
# Create
# ==================================================================================================
# 
#
def dataframe(inp, test=False, autotranspose=False, verbose=None):
    """ 
    Converts multidimensional objects into dataframes.
    Dictionaries and Tuples are interpreted column-wise, Lists and Counters by rows.
    (Exception: A list of Series is also interpreted column-wise, like a tuple of Series.)
    * test:           Do additional test if the shape is ok
    * autotranspose:  False => no autotranspose
                      n     => If the result is n times wider than long, it is transposed.  
    """
    
    if verbose is None:
        verbose = Config.get('VERBOSE')   
        
    if isinstance(inp, pd.Series):
        return pd.DataFrame(inp)
    
    def do_test(inp, result, test, verbose, inp_rtype, inp_shape, gedreht):
        if verbose:
            print('rotated='+str(gedreht) + ' Output rtype=' + str(bpy.rtype(result)), 'shape=' + str(bpy.shape(result)))
        if not test:
            return True
        
        # Nur die ersten beiden Dimensionen interessieren
        inp_shape = inp_shape[:2]
        
        # Erste Dimension ggf. ergänzen
        if len(inp_shape) == 1:
            inp_shape = (1,inp_shape[0])          
            
        # spaltenweise interpretiert?
        elif gedreht:
            inp_shape = (inp_shape[1],inp_shape[0])

        
        if inp_shape != bpy.shape(result)[:2]:
            if verbose:
                print('Shape does not fit')            
            return False
        return True
    
    
    # Spaltennamen
    def cols_benennen(result):
        if result.shape[1] > 52: # zu breit, kann man nicht umbenennen
            return result
        # gibt es Duplikate in den Spaltennamen?                       oder sind die Spalten rein numerisch?
        if (len(list(result.columns))  !=  len(set(result.columns)))   or   isinstance(result.columns, pd.RangeIndex):     
            result.columns = list('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')[0:result.shape[1]] 
        return result
    
    # Analyse
    inp_rtype = bpy.rtype(inp)
    gedreht = False # normale Orientierung
    try:
        inp_shape = bpy.shape(inp) 
    except:
        inp_shape = (-77,-77)
    if verbose:
        print('Input rtype=' + str(inp_rtype), 'shape=' + str(inp_shape))
    
    # tuple of Series oder list of Series: 
    # in dict wandeln und gedreht vormerken 
    if len(inp_rtype)>=2  and  inp_rtype[0] in ['list','tuple']  and  inp_rtype[1] == 'Series':
        sp = tuple(a.name for a in inp) 
        if (len(list(sp))  !=  len(set(sp))): # dups?    
            sp = tuple('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')[0:len(inp)]        
        inp = dict(zip(sp, inp)) 
        gedreht = True
        
    # tuple: 
    # in dict wandeln und gedreht vormerken    
    elif inp_rtype[0] == 'tuple':
        sp = tuple('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')[0:len(inp)] 
        inp = dict(zip(sp, inp))  
        gedreht = True
        
    # dict: 
    # gedreht vormerken   
    elif inp_rtype[0] in ['dict']:
        gedreht = True        
    
    # 0 Dimensionen: Leeres DataFrame
    if len(inp_shape) == 0:
        result = pd.DataFrame()
        assert do_test(inp, result, test=test, verbose=verbose, inp_rtype=inp_rtype, inp_shape=inp_shape, gedreht=gedreht)
        return result
    
    # 1 Dimensionen: Eine Zeile
    if len(inp_shape) == 1:
        result = pd.DataFrame([inp])
        assert do_test([inp], result, test=test, verbose=verbose,inp_rtype=inp_rtype, inp_shape=inp_shape, gedreht=gedreht)
        if autotranspose  and  result.shape[0]*autotranspose < result.shape[1]:  # Hundertmal breiter als lang: transpose
            result = result.transpose()    
        result = cols_benennen(result)             
        return result    
    
    if isinstance(inp, dict): 
        result = pd.DataFrame.from_records(inp, columns=inp.keys())
    else:
        result = pd.DataFrame.from_records(inp)

    assert do_test(inp, result, test=test, verbose=verbose, inp_rtype=inp_rtype, inp_shape=inp_shape, gedreht=gedreht)
    if autotranspose  and  result.shape[0]*autotranspose < result.shape[1]: # Hundertmal breiter als lang: transpose
        result = result.transpose()
    result = cols_benennen(result)        
    return result



# ==================================================================================================
# Columns
# ==================================================================================================

def drop_cols(df, colnames):
    '''
    Drops a column or a list of columns.
    Does not throw an error if the column does not exist.
    '''
    
    if type(colnames) is str:
        colnames = [colnames] #let the command take a string or list
        
    result = df.copy()
    for c in colnames:
        while c in result.columns:
            result = result.drop(c, axis=1)        
    return result



def rename_col(df, name_from, name_to):
    '''
    Renames a column of a DataFrame.
    If you try to rename a column again, no error is thrown (better for the workflow in jupyter notebooks).
    '''
    if name_from == name_to:
        return df
    if (name_to in df.columns) and (not name_from in df.columns):
        return df # if you try to rename a column again, no error is thrown.
    if name_to in df.columns:
        raise ValueError( name_to + ' already exists!')
    if not name_from in df.columns:
        raise ValueError( name_from + ' does not exist!')        
    return df.rename( columns = { name_from: name_to } )    



def move_cols( df, colnames=None, to=0 ):
    ''' 
    Reorders the columns of a DataFrame. 
    The specified columns are moved to a numerical position or behind a named column.
    If no arguments are given, sorts the columns lexicographically.
     * colnames: String or list of column names
     * to:       0: move to the front
                -1: move to the back
               <i>: move to position i (buggy)
         <colname>: move behind a specific column
    '''   
    if colnames is None:
        return df.reindex(sorted(df.columns), axis=1)  
    if type(colnames) is str:
        colnames = [colnames] #let the command take a string or list
        
    if type(to) is str:       
        to = list(df.columns).index(to) + 1
        
    colnames_sort  = [c for c in colnames   if c in df.columns]      
    colnames_other = [c for c in df.columns if c not in colnames]

    
    if to==0:
        return df[ colnames_sort + colnames_other ]
    elif to==-1:    
        return df[ colnames_other + colnames_sort] 
    else:
        cols_start = [c for c in list(df.columns)[:to]  if c not in colnames_sort]
        cols_rest  = [c for c in df.columns             if c not in cols_start and c not in colnames_sort]   
        #print(cols_start)
        #print(colnames_sort)        
        #print(cols_rest)            
        return df[ cols_start + colnames_sort + cols_rest] 

    
        

        
    

# ==================================================================================================
# update_col
# ==================================================================================================


def update_col(df_to, df_from, on=[], left_on=[], right_on=[], col='', col_rename='', col_score='', func='', cond='', keep='', return_mask=False, verbose=None):
    '''
    Transfers one column of data from one dataframe to another dataframe.
    Unlike a simple merge, the index and the dtypes are retained. 
    Handles dups and conditions. Verbose if wanted.
    
    df_to:        Dataframe to be changed
    df_from:      Dataframe that contains the new data. Does not have to be duplicate-free.
    on:           Column name or list of column names whose values must match 
    left_on:      Different for left and right if necessary
    right_on:     Different for left and right if necessary    
    col:          Name of the column to be transmitted. Must exist in df_from. 
                  If not present in df_to, it will be appended. 
                  If already present, matching values will be overwritten.
    col_rename:   New name for col, if specified. If already present, matching values will be overwritten and not matching values are preserved.      
    col_score:     Name of a score column used for picking rows from df_from with maximum score. This is useful to avoid conflicting values in case of dups in df_from.
                  If empty: No dup avoidance by rank.
    func:         Name of the function used for dup avoidance. E.g. 'max'. If empty: No dup avoidance by func.
                  Might be slow, use col_score if possible.   
    cond:         Empty, 'min','max' or 'null'. 
                  'min','max': Only write if the new value is smaller / larger than the existing value
                  'null'     : Only write if there is no existing value in df_to (it's Null or empty string) 
    keep:         Should the original value be kept in a separate column? 
                  The string keep is used as suffix for this column.
                  NaN if record is unchanged!  
    return_mask:  If True, the result is a tuple of the resulting dataframe plus a mask masking the affected rows
    verbose:    Messages on / off
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
    
    # col_rename
    if col_rename=='':
        col_rename = col
    
    # keep ggf. korrigieren
    if not col in df_to.columns:
        keep = ''    
    
    # auf left_on und right_on umstellen
    if on:
        left_on  = on
        right_on = on            
        
    # Listen erzwingen
    if left_on  and not isinstance(left_on, list):
        left_on = [left_on]
        
    if right_on and not isinstance(right_on, list):
        right_on = [right_on] 
        
        
    # df_from vorbereiten
    if func:
        anz = df_from.shape[0]
        df_from = df_from[  right_on+[col]  ].copy()
        gruppiert = df_from.groupby(right_on)
        df_from = gruppiert.agg(
            col_max = (col, func ),                          
        ).reset_index()   
        df_from = rename_col(df_from, 'col_max', col)
        if (df_from.shape[0] == anz)  and  verbose:
            print('update_col:','func' ,func, 'applied, but it was pointless!')
        elif (df_from.shape[0] != anz)  and  verbose:
            print('update_col:','func' ,func, 'applied,', anz - df_from.shape[0], 'records less!')   
            
    elif col_score:
        anz = df_from.shape[0]
        df_from = rank(df_from, col_score=col_score, cols_group=right_on, on_conflict='first', verbose=False)[ right_on+[col] ].copy()
        if (df_from.shape[0] == anz)  and  verbose:
            print('update_col:','col_score' ,col_score, 'applied, but it was pointless!')
        elif (df_from.shape[0] != anz)  and  verbose:
            print('update_col:','col_score' ,col_score, 'applied,', anz - df_from.shape[0], 'records less!')           
        
    else:
        df_from = df_from[ right_on+[col] ].copy()
    
    # umbenennen wie left bzw. df_to
    df_from.columns = left_on+[col_rename]
        
    # Index merken
    df_to = df_to.copy()
    df_to['copy_index'] = df_to.index
    copy_index_name = df_to.index.name
    
    #merge
    result = df_to.merge(df_from, on=left_on, how='left', suffixes= ('', '_new')) 
    
    # Index restaurieren
    result = result.set_index('copy_index')
    result.index.name = copy_index_name

    # keep
    if keep:
        result[col_rename+keep] = result[col_rename]  
    
    # Zielspalte existiert schon
    if col_rename in df_to.columns: 
        col_new = col_rename + '_new'
        # result[col_rename] ist der alte Wert
        # result[col_new]    ist der neue Wert        
        if cond == 'min':
            mask = result[col_new].notnull()   &   (result[col_new] < result[col_rename]) 
        elif cond == 'max':
            mask = result[col_new].notnull()   &   (result[col_new] > result[col_rename]) 
        elif cond == 'null':
            mask = result[col_new].notnull()   &   (  result[col_rename].isnull()   |   (result[col_rename]=='')   )        
        else:         
            mask = result[col_new].notnull() 
        if verbose:
            print('update_col:', result[mask].shape[0], 'cells written into existing column')
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)   
            result.loc[mask, col_rename] = result.loc[mask, col_new]
        result = copy_datatype( result, df_to )    # datatypes restaurieren               
        #result[col_rename] = copy_datatype( result[col_rename], df_to[col_rename] )    # datatype restaurieren
        result = drop_cols(result,[col_new])
        result_mask = mask.copy()
    
    # Die Spalte ist neu
    else:
        #print(list(df_from.columns))
        result             = copy_datatype( result,             df_to        )           # datatypes restaurieren        
        result[col_rename] = copy_datatype( result[col_rename], df_from[col_rename] )    # datatype übernehmen    
        result_mask = result[col_rename].notnull()
        if verbose:
            print('update_col:', result[result_mask].shape[0], 'cells written into new column')        
    
    if keep:
        mask = result[col_rename+keep] == result[col_rename]
        result.loc[mask,col_rename+keep] = np.NaN
        
    if df_to.shape[0] != result.shape[0]:
        print('update_col:','WARNING: df_from identifier not unique.','I call this again with func="max"')
        return update_col(df_to, df_from, left_on=left_on, right_on=right_on, col=col, col_rename=col_rename, func='max', keep=keep, verbose=verbose)
    
    if return_mask:
        return result, result_mask
    return result

    
    
def write_empty_col(df,col_name, content):
    '''
    Writes empty iterables into a column and sets datatype.
    * content: Can be 'list','set','dict','defaultdict', 'string' or any empty iterable.
    Example: write_empty_col(df,'mycol','list') writes empty lists in mycol.
    '''
    result = df.copy()
    if content == 'string':
        result[col_name] = ''
        result[col_name] = result[col_name].astype('string')      
        return result
    if content == 'list':
        content = list()
    elif content == 'set':
        content = set()     
    elif content == 'dict':
        content = dict()   
    elif content == 'defaultdict':
        content = defaultdict(list)

    result[col_name] = [ content for x in range(len(df.index))]
    result[col_name] = result[col_name].astype('object')
    return result   
    
    
    
def copy_datatype(data_to, data_from):
    """
    Copies the dtypes from data_from to data_to. 
    Usable for Series and DataFrames.
    When applied on a DataFrame, it's applied to all column names that match.
    """
    
    result = data_to.copy()
    
    if isinstance(data_to, pd.DataFrame)   and  isinstance(data_from, pd.DataFrame):
        for c in result.columns:
            if c in data_from.columns:
                try:
                    result[c] = result[c].astype( data_from[c].dtypes.name )  
                except:
                    pass
        return result
                
    elif isinstance(data_to, pd.Series)   and  isinstance(data_from, pd.Series):
        #print(data_from.dtypes.name)
        result = result.astype(data_from.dtypes.name)
        return result
    
    else: 
        assert 'Wrong type'



    

    

    

# ==================================================================================================
# Index
# ==================================================================================================
    

def reset_index(df, keep_as=None):
    '''
    Creates a new, unnamed index.
    * keep_as: If keep_as is given, the old index is preserved as a row with this name.
    Otherwise the old index is dropped.
    '''
    
    if len(df.index.names) != 1:
        raise ValueError('index must be 1dim')
    
    if keep_as:
        df.index.names = [keep_as]    
        result = df.reset_index()
        result.index.names = [None]  
        result.columns.name = None
        return result
    
    else:
        df.index.names = [None]   
        result = df.reset_index(drop=True)
        result.columns.name = None
        return result        

    

# benennt den Index um
def rename_index(df, soll):
    '''
    Renames the index
    '''
    df.index.names = [soll]   
    return df




def drop_multiindex(df, verbose=None):
    '''
    Converts any MultiIndex to normal columns and resets the index. 
    Works with MultiIndex in Series or DataFrames, in rows and in columns.
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
    
    # Series mit MultiIndex
    if isinstance(df, pd.Series)   and   (df.index.nlevels > 1): 
        if verbose:
            print('Series with MultiIndex')
        return df.reset_index()    
    
    # Series ohne MultiIndex
    if isinstance(df, pd.Series): 
        if verbose:
            print('Series without MultiIndex')        
        return df    
    
    # DataFrame ohne MultiIndex: 
    if (df.columns.nlevels <= 1)   and   (df.index.nlevels <= 1):
        if verbose:
            print('DataFrame without MultiIndex')             
        return df
    
    # DataFrame mit Zeilen-MultiIndex: 
    if (df.columns.nlevels <= 1)   and   (df.index.nlevels > 1):
        if verbose:
            print('DataFrame with Row-MultiIndex')         
        return df.reset_index()   
    
    # DataFrame mit Spalten-MultiIndex: 
    if (df.columns.nlevels > 1)   and   (df.index.nlevels <= 1):
        if verbose:
            print('DataFrame with Column-MultiIndex')            
        result = df.copy()
        result.columns = ['{}_{}'.format(col[0], col[1]) for col in result.columns]
        return result
    
    # DataFrame mit Zeilen- und Spalten-MultiIndex: 
    if (df.columns.nlevels > 1)   and   (df.index.nlevels > 1):
        if verbose:
            print('DataFrame with Row- and Column-MultiIndex')          
        result = df.copy()
        result.columns = ['{}_{}'.format(col[0], col[1]) for col in result.columns]
        return result.reset_index()   
   

    #print([(col[0], col[1]) for col in df.columns])

    
    
    
# Liefert einen Dataframe mit dem angegebenen Index.
# Wenn als Index ein Dataframe angegeben wird, wird dessen Index verwendet.
# Bsp:
# force_index(lemmas_w, ['tagZ','lemma_lower'])
# force_index(lemmas_w, stat)
#
def force_index(df, soll):
    
    import warnings
    warnings.warn("deprecated, use update_col for the old usecase", DeprecationWarning)    
    
    # ist und soll festlegen
    index_soll = soll
    if type(soll)== pd.DataFrame:
        index_soll = soll.index.names
    index_ist = df.index.names
    
    if index_ist == index_soll:
        return df
    else:
        result = reset_index(df)
        return result.set_index( index_soll)

# ==================================================================================================
# Rows
# ==================================================================================================


def drop_rows(df, mask, verbose=None):
    '''
    Drops rows identified by a binary mask.
    * verbose: True if you want to print how many rows are droped.
    (If you want to delete the rows to a trash, use move_rows.)
    '''

    if verbose is None:
        verbose = Config.get('VERBOSE') 
        
    # damit die Negation funktioniert
    try:
        mask = mask.fillna(False)
    except:
        mask = np.nan_to_num(mask)    
        
    # verbose Statusmeldung
    if verbose:
        anz = df[mask].shape[0]
        if anz == 0: 
            print('No rows deleted')        
        else: 
            print('Delete', anz, 'rows from', df.shape[0])    
        
    return df[~mask].copy() 




def move_rows(df_from, df_to=None, mask=None, msg='', msgcol='msg', verbose=None, zähler=[0]):
    '''
    Moves rows identified by a binary mask from one dataframe to another (e.g. into a trash).
    Returns two DataFrames.
    The target dataframe gets an additional message column by standard (to identify why the rows were moved).
    If you don't give a message, move_rows will generate one: just the count of usage of the function.
     
    * df_from:   Origin DataFrame
    * df_to:     Target DataFrame or None
    * mask:      Binary mask, identifies the rows to be deleted
    * msg:       All moved rows are marked with this message, it's written in the message column of the target dataframe
    * msgcol:    Name of the message column. Set msg or msgcol to None if no message wanted.
    * verbose:   True if you want to print how many rows are moved.
    
    Examples:
    df, df_trash = move_rows( df, mask )           # move rows away (and create a new trash). 
                                                   # The second argument is detected as mask, not as df_to.
    df, df_trash = move_rows( df, df_trash, mask ) # move rows away (into the existing trash)   
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE') 
        
    if not isinstance(df_from, pd.DataFrame):
        raise ValueError('df_from must be a DataFrame')         
        
    # df_to enthält mask    
    if isinstance(df_to, pd.Series):
        mask  = df_to
        df_to = None
        
    # msg   
    if not df_to is None:
        zähler[0]+=1   
    if msg=='': 
        msg = str(zähler[0])        
        
    # damit die Negation funktioniert
    try:
        mask = mask.fillna(False)
    except:
        mask = np.nan_to_num(mask) 
        
    # anz    
    anz = df_from[mask].shape[0]         
       
    # verbose Statusmeldung
    if verbose:
        if anz == 0: 
            print('No rows moved')               
        else: 
            print('Move', anz, 'rows from', df_from.shape[0])     
            
    # anz == 0    
    if anz == 0 and not df_to is None:      
        return df_from, df_to       
    if anz == 0 and df_to is None:     
        return df_from, pd.DataFrame()        
    
    # Löschen bestehendem trash    
    if not df_to is None:     
        t = df_from[mask].copy()
        if msgcol and not msg is None:        
            t[msgcol] = msg  # kennzeichnen   
            t[msgcol] = t[msgcol].astype('string')
        r1 = df_from[~mask].copy()
        r2 = pd.concat([df_to, t])
        r2 = copy_datatype(r2, r1)    # falls neue Spalten dazugekommen sind         
        return r1, r2   
    
    # Löschen mit neuem trash    
    if df_to is None:     
        t = df_from[mask].copy()
        if msgcol and not msg is None:
            t[msgcol] = msg # kennzeichnen  
            t[msgcol] = t[msgcol].astype('string')
        return df_from[~mask].copy() , t
        
    return "ERROR"            
    









def add_rows(df_main, df_add, only_new=None, reindex=True, assert_subset=False, verbose=None):
    '''
    Like concat, with additional features only_new and verbose.
    * df_main:       The new rows are added to the end of this dataframe.
    * df_add:        Rows to add. The dtypes are adapted to those of df_main.
                     Series or list are also accepted for appending.
    * only_new:      Avoid duplicates by setting this to a list of column names.
                     This combination must contain new content.
                     A single column name as string works the same way.
                     Or set only_new=True, this will avoid duplicate row indexes.
    * reindex:       Will the result get a fresh index without dups?
    * assert_subset: Check if all columns in df_add exist in df_main already?
    * verbose:       Print status messages how many rows affected
    '''
    
    if not type(df_main) is pd.DataFrame:
        df_main = dataframe(df_main, verbose=False)
        
    if not type(df_add) is pd.DataFrame:
        df_add = dataframe(df_add, verbose=False)
    
    if verbose is None:
        verbose = Config.get('VERBOSE')   
        
    if assert_subset: 
        assert set(df_add.columns) <= set(df_main.columns)        
    
    # only_new
    if only_new:
        
        #let the command take a string as well 
        if isinstance(only_new,str):  
            only_new = [only_new]  
            
        # only_new is list >> recursiv call
        if isinstance(only_new,list):  
            mask = ~isin(df_add, df_main, on=only_new)  
            if verbose:
                print(df_add.shape[0]-df_add[mask].shape[0], 'rows not attached')        
            return add_rows(df_main, df_add[mask], reindex=reindex, verbose=verbose)
        
        # only_new=True >> recursiv call
        if only_new==True:
            mask = ~df_add.index.isin(df_main.index)   
            if verbose:
                print(df_add.shape[0]-df_add[mask].shape[0], 'rows not attached')        
            return add_rows(df_main, df_add[mask], reindex=reindex, verbose=verbose)
            
        raise ValueError('only_new must be str, list or True') 
        
    # dtypes anpassen
    df_add = copy_datatype(df_add, df_main)     
    
    # result  
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)     
        result = pd.concat([df_main, df_add], ignore_index=reindex) 
    
    # Statusmeldung
    if verbose:
        if result.index.is_unique:        
            print(df_add.shape[0], 'rows added, now a total of', result.shape[0])
        else:
            print(df_add.shape[0], 'rows added, now a total of', result.shape[0], 'Warning: Index is not unique!')            

            
    
    return result



def quicksample(data, size):
    '''
    Returns a subset of the input with the given size,
    including the first and the last rows.
    Works with DataFrame or Series.
    '''
    if size <= 1:
        return data.head(size)
    if data.shape[0] <= size:
        return data    

    if size == 2:
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)   
            result = pd.concat( [data.head(1),data.tail(1)] )
        return result
    
    if size < 10:
        n = 1
    elif size < 100:
        n = 2        
    else:
        n = 3        
    r1 = data.head(n)
    r2 = data.iloc[n:-n].sample(size-n-n)
    r3 = data.tail(n)    
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)      
        result = pd.concat([r1,r2,r3])
    return result 


    


# ==================================================================================================
# first_valid_value und last_valid_value
# ==================================================================================================
 

    
def first_valid_value(series):
    '''
    Returns the first not-nan values of a Series. 
    '''
    try:
        idx = series.first_valid_index()
        if idx is None:
            return None
        result = series.loc[idx]
        if isinstance(result, pd.Series): # das liegt an nonunique Index
            return result.iloc[0]
        else:
            return result    
    except:
        return None        
    


def last_valid_value(series):
    '''
    Returns the last not-nan values of a Series. 
    '''    
    try:
        idx = series.last_valid_index()
        if idx is None:
            return None
        result = series.loc[idx]
        if isinstance(result, pd.Series): # das liegt an nonunique Index
            return result.iloc[-1]
        else:
            return result    
    except:
        return None     
    




# ==================================================================================================
# list
# ==================================================================================================
 

#
def find_in_list( df, col_list_of_strings, searchstring ):
    '''
    Searches a column with a list of strings.
    Returns a binary mask for the rows containing the searchstring in the list.    
    '''
    zusatzfelder = pd.DataFrame(  df[col_list_of_strings].explode()  )
    mask_z  = (  zusatzfelder[col_list_of_strings] == searchstring  )
    auswahl = zusatzfelder[mask_z]
    mask    = df.index.isin(auswahl.index)   
    return mask   




def apply_on_elements(series, funktion):
    '''
    Applies a function to all elements of a Series of lists.
    Example:
    df = pak.people()
    df['history2'] = pak.apply_on_elements(df.history, lambda x: x+'2' if x==x else '')    
    Also works with sets.
    '''
    return series.explode().apply(funktion).groupby(level=0).apply(list)




def list_to_string(series, sep=','):
    '''
    Converts a Series of lists of strings into a Series of strings.
    * sep: The separator, default is ','
    Example:
    df = pak.people()
    df['history2'] = pak.list_to_string(df.history)    
    '''

    def try_join(l):
        if not l:
            return ''
        try:
            return sep.join(map(str, l))
        except TypeError:
            return str(l)

    result = [try_join(l) for l in series]
    result = pd.Series(result).astype('string')
    return result
    
    


####################################################################################################
# ..................................................................................................
# Mehrere Dataframes 
# ..................................................................................................
####################################################################################################



# ==================================================================================================
# isin
#   
def isin( df1, df2, on=[], left_on=[], right_on=[] ):
    '''
    isin over several columns. 
    Returns a mask for df1: The rows of df1 that match the ones in df2 in the specified columns.
    
    '''
    if on:
        left_on  = on
        right_on = on        
        
    i1 = df1.set_index(left_on).index
    i2 = df2.set_index(right_on).index    
    result = i1.isin(i2)   
    result = pd.Series(result).to_numpy(na_value=False)
    return result   



#################################################################################################
# ...............................................................................................
# Series
# ...............................................................................................
#################################################################################################


def repeat(content, size):
    '''
    Creates a Series with defined size by repeating a list 
    '''
    s = int(size / len(content))
    result = pd.Series(content).repeat(s)
    if size % len(content) == 0:
        return result.reset_index(drop=True)
    else:
        return pd.concat([result,pd.Series(content)]).head(size).reset_index(drop=True)







import collections, warnings

from functools   import partial  

import numpy  as np
import pandas as pd

from pandas.api.types import is_string_dtype, is_numeric_dtype

import bpyth  as bpy

from .type_info    import type_info
from .values_info  import values_info

from .config       import Config
from .pandas       import dataframe, reset_index, drop_cols, rename_col, move_cols, quicksample, first_valid_value, last_valid_value
from .aggregate    import group_and_agg, top_values, top_values_count, most_freq_elt
from .scale        import scale
from .rank         import rank_without_group


    
try:
    from termcolor import colored    
except ImportError:
    pass


import matplotlib.pyplot as plt
#sns.set()


try:
    import seaborn as sns
except ImportError:
    print('seaborn not importable')  
    






# ==================================================================================================
# Load und save (gehört eigentlich nicht hierher, muss aber wg. changedatatype
# ==================================================================================================
# 
#
def load_pickle( filename, resetindex='AUTO', changedatatype=False, verbose=None ): 
    '''
    Convenient function to load a DataFrame from pickle file.
    Optional optimisation of datatypes. Verbose if wanted.
    resetindex = True:    Force reset_index
    resetindex = False:   No reset_index    
    resetindex = 'Auto':  (Standard) Automatic     
    changedatatype:       Should the datatypes get optimized?
    verbose:              True if messages are wanted.
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')  
        
    result = bpy.load_pickle(filename)
    if resetindex == True:
        result = result.reset_index()
    elif resetindex == 'AUTO':
        result = result.reset_index()    
        result = drop_cols(result, 'index')       
    
    if changedatatype:     
        result = change_datatype(result, verbose=verbose)

    if verbose:
        print(result.shape[0], 'rows loaded')  
        
    return result




def dump_pickle( df, filename, changedatatype=True, verbose=None ): 
    '''
    Convenient function to save a DataFrame to a pickle file.
    Optional optimisation of datatypes. Verbose if wanted.
    changedatatype:       Should the datatypes get optimized?
    verbose:              True if messages are wanted.    
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')  
        
    if changedatatype:       
        df = change_datatype(df, verbose=False)
    bpy.dump_pickle(df,filename)

    
    
    
#################################################################################
# ...............................................................................
# Spalten auf datatype untersuchen
# ...............................................................................
#################################################################################    
    

    
def mem_usage(data):
    """Returns the memory consumption of a Series or a DataFrame"""
    
    if isinstance(data, pd.Series): 
        result = data.memory_usage(index=False, deep=True)
        return bpy.human_readable_bytes(result)    
    
    elif isinstance(data, pd.DataFrame):
        result = data.memory_usage(index=False, deep=True).sum()
        return bpy.human_readable_bytes(result)    
        
    else:
        assert 'ERROR'    




# Alle class_info einer Series oder eines Index   
def analyse_datatype(data):
    """ 
    Returns a dict with info about the datatypes of a Series or Index and it's content
    """
    
    # Aufruf mit Index
    if isinstance(data, pd.Index): 
        series = data.to_series()
        series.name = '__index__'
        return analyse_datatype(series)    

    info = type_info(data)
    result = {
        'col_name': data.name,
        'datatype_instance': info.name_instance,  
        'datatype': info.name,        
        'datatype_short': info.name_short,        
        'is_numeric': is_numeric_dtype(data),                
        'is_string': is_string_dtype(data),  
        'is_hashable': info.is_hashable,
        'nan_allowed': info.nan_allowed,   
        'mem_usage': mem_usage(data),        
    }
        
    return result



# Alle class_info eines DataFrame
def analyse_datatypes(df, with_index=True):
    """
    Returns info about the datatypes and the mem_usage of the columns of a DataFrame.  
    """
    if isinstance(df, pd.Series): 
        return dataframe( analyse_datatype(df), verbose=False ) 
    
    # Kleine Probe reicht
    if df.shape[0] > 10:
        return analyse_datatypes(quicksample(df, 10), with_index=with_index)
    
    data  = [] 
    if with_index:
        data += [ analyse_datatype(df.index)]
    data     += [ analyse_datatype(df[col]) for col in df ]

    result = dataframe(data, verbose=False)

    # Zeilen-, Spalten- und Tabellenname
    result = result.rename_axis('col_no')    
    result = result.rename_axis('', axis='columns')
    return result



# Spaltennamen, die bestimmte Kriterien erfüllen
def col_names(df, only='', without='XXXXXX', as_list=True, query=None, sort=False ):
    """ 
    Selects column names based on analyse_cols. Useful to apply a method to specific columns of a DataFrame.
    * only:     Only column names whose datatype begins like this.
    * without:  Without column names whose datatype starts like this
    * as_list:  Output result as list (otherwise as DataFrame, useful for development and control)
    * query:    additional conditions
    * sort:     sorted by nunique
    datatypes are taken from analyse_datatypes, field datatype_short or datatype_instance.

    Example: treat all str columns with fillna
    cols = col_names(df, only='str', query='nnan > 0')
    df[cols] = df[cols].fillna('')        
    """
    if sort:
        df = sort_cols_by_nunique(df)
        
    if query or not as_list:
        info = analyse_cols(     df, with_index=False)      # komplette Analyse holen
    else:
        info = analyse_datatypes(df, with_index=False)      # nur datatypes analysieren      
    
    mask1 =  info.datatype_short.str.startswith(only)      |    info.datatype_instance.str.startswith(only)     |    info.datatype.str.startswith(only)
    mask2 = ~info.datatype_short.str.startswith(without)   &   ~info.datatype_instance.str.startswith(without)  &   ~info.datatype.str.startswith(without)
    result = info[mask1 & mask2]
    
    # Soll noch eine Query angewendet werden?
    if query:
        result = result.query(query)
    
    # Kompletten Dataframe zurückgeben?
    if not as_list:
        return result
    
    # als Liste zurückgeben
    return list(result.col_name)
    

    

    
    
def change_datatype(data, search=None, verbose=None, msg='', category_maxsize=-1, nanless_ints=False):
    """ 
    Converts the datatypes of a DataFrame or a Series.
    If used with a Series:    
    Similar behavior as pandas astype. But it also accepts
    sloppy class names like type_info knows.
    If no target datatype is specified, it will be selected automatically.
    If used with a DataFrame:
    Converts all datatypes automatically.                      

    * category_maxsize: How big can a category get to be suggested as datatype_suggest?
    * nanless_ints: Are numpy's integer classes (that don't know NaN) suggested as datatype_suggest?    
    """

    if verbose is None:
        verbose = Config.get('VERBOSE')   

    
    # Series
    if isinstance(data, pd.Series): 
                
        # vorgegebener Datatype
        if search: 
            i = type_info(search)              
            return data.astype(i.name_short)
            
        # vollautomatisch    
        else: 
            search = values_info(data, category_maxsize=category_maxsize).datatype_suggest  
            if search:
                if verbose:
                    print( '{:<20s} --> {:<10s}'.format(data.name, search) )                    
                return change_datatype(data, search=search, verbose=verbose, msg=msg)  
            else:
                return data

    # DataFrame
    elif isinstance(data, pd.DataFrame)   and not search:
            
        if verbose:
            print('change_datatype', msg )        
        result = data.apply(change_datatype, verbose=verbose   )
        if verbose:
            print('change_datatype','before:', mem_usage(data),'after:',mem_usage(result))
            print()
        return result
    
    else:
        assert 'ERROR'
    
# Synonym    
change_datatypes = change_datatype    
    
    
    
#################################################################################
# ...............................................................................
# Spalten auf values untersuchen
# ...............................................................................
#################################################################################     
    
    
def analyse_values(data, as_list=False, as_dict=False, sort=False, with_index=True, nanless_ints=False):
    """ 
    Returns statistical data for a DataFrame, a Series or an Index     
    """
    
    # Aufruf mit Index
    if isinstance(data, pd.Index): 
        series = data.to_series()
        series.name = '__index__'
        return analyse_values(series, as_list=as_list, as_dict=as_dict, nanless_ints=True)
    
    
    # Aufruf mit DataFrame
    if isinstance(data, pd.DataFrame): 
        info  = []
        if with_index:
            info += [ analyse_values(data.index, as_list=True) ]         
        info     += [ analyse_values(data[col],  as_list=True) for col in data ] 
        result = pd.DataFrame.from_records(info)

        # Zeilen-, Spalten- und Tabellenname
        result.columns = analyse_values(pd.Series(1)).index  # Sample abholen, Series ist dabei egal
        result = result.rename_axis('col_no')    
        result = result.rename_axis('', axis='columns')

        if sort:
            result = result.sort_values(['nunique','col_name'], ascending=[False,True])    
        return result        
    
        
    # Aufruf mit Series  
    assert isinstance(data, pd.Series)   
    info = values_info(data, nanless_ints=nanless_ints)
    result = [
        data.name,
        info.ntypes,        
        info.nunique,            
        info.nnan,
        info.ndups, 
        info.n,   
        info.vmin,      
        info.vmean,        
        info.vmedian,
        info.vmax,      
        info.vsum,            
        info.datatype_suggest,
    ]
    
    # Rückgabe als Liste
    if as_list:
        return result
    
    # In DataFrame wandeln
    result = pd.DataFrame(result)
    result['analyse'] = pd.Series(['col_name','ntypes', 'nunique','nnan','ndups','n','vmin','vmean','vmedian','vmax','vsum','datatype_suggest'])
    result = result.set_index('analyse')
    
    # Rückgabe als dict
    if as_dict:
        result = result.to_dict()[0]
        del result['col_name']
        return result    
    
    # Rückgabe als DataFrame
    return result    
    
    
    

# ==================================================================================================
# analyse_cols
# ==================================================================================================

def analyse_cols(df, sort=False, with_index=True):
    """ 
    Describes the datatypes and the content of a DataFrame.
    Merged info from analyse_datatypes and analyse_values.
    """   
        
    info1  = analyse_datatypes(df, with_index=with_index)
    info2  = analyse_values(   df, with_index=with_index)
    result = pd.merge(info1, info2)    
    result = move_cols(result, ['col_name','datatype_instance','datatype','datatype_short','datatype_suggest'])

    if sort:
        result = result.sort_values(['nunique','col_name'], ascending=[False,True])    
    return result    

            
        
    
##############################################################################################
# ............................................................................................
# NaN, NA, None 
# ............................................................................................
##############################################################################################
    

# ersetzt analyse_nans und nan_anz
def nnan(data, all=False): #, sum=False):
    """ 
    Count NaNs in Series or DataFrames.          
    For a DataFrame: Returns a list of all fields and the number of NaNs. 
    For a Series:  Returns the number of NaNs.    
    * all: Should the result include also the columns without NaNs (for a DataFrame)?
    * sum: Should the result include only the total sum (for a DataFrame)? Identical with nnan(df).sum()          
    """
    result = data.isnull().sum()
    if isinstance(data, pd.Series) or all:
        return result
    
    return result[result!=0]

  
    

# ersetzt assert_no_nans    
def any_nan(data, without=None):
    """ 
    Are there NaNs? Returns True or False.
    Works for Series or DataFrame.
    assert not any_nan(df) ensures that a dataframe does not contain NaNs.
    * without contains a field name or a list of field names that will be excluded.        
    """    
    if not without:
        return data.isnull().values.any()
    
    return drop_cols(data, without).isnull().values.any()
        



    
def nan_rows(df, col=''):   
    """ 
    Returns the rows of a DataFrame that are NaN in the specified column.  
    * col not specified: Returns all rows with NaNs in any column
    * col=0:             Returns all rows with NaN in the FIRST column with NaNs. 
    * col='city':        Returns all rows with NaNs in the column 'city'
    """
    
    # no column specified: 
    if col=='':
        return df[df.isna().any(axis=1)]
    
    # select the first column with NaNs
    elif col==0:
        col = nnan(df).head(1).index.to_list()[0]
    
    # return rows
    mask = df[col].isnull()
    return df[mask]
    
    


 
    
    
    
    

#################################################################################################
# ...............................................................................................
# Series 
# ...............................................................................................
#################################################################################################





# ==============================================================================================
# verteilung
# ==============================================================================================

#def verteilung(series, style=None, quantile=1, stat=True):
#    '''
#    Veraltet. Verwende analyse_freqs stattdessen.
#    Liefert Informationen über die Verteilung einer Series
#    style:       ('key','top','plot')
#    quantile:  wird nur für den Plot verwendet und schneidet ihn ab
#    basiert auf seaborn für die grafische Darstellung bzw. auf countgrid für die textuelle Darstellung
#    Beispiel siehe Pandas/Analyse
#    '''
#    warnings.warn('Veraltet. Verwende analyse_freqs stattdessen.')
#    from matplotlib import pyplot as plt
#    
#    # Mini-Statistik
#    s = analyse_values(series, as_dict=True)
#    if stat:
#        print(s)   
#        print()
#    
#    # automatische style-Auswahl
#    if not style  and  is_string_dtype(series):
#        style = 'key'  
#    if not style  and  not s:
#        style = 'key'      
#    if not style  and  (int(s['nunique'])<=7):
#        style = 'key'   
#    if not style:  
#        style = 'top'         
#    
#    # textuell
#    if (style == 'key'): 
#        return countgrid(series, sort=False)
#    
#    # textuell
#    elif (style == 'top'): 
#        return countgrid(series, sort=True)    
#    
#    
#    # grafisch
#    else:
#        mask = (series <= series.quantile(quantile))   &   (series >= series.quantile(1-quantile)) 
#        try:
#            plt.figure(figsize=(16, 4))
#            return seaborn.histplot(series[mask])
#        except RuntimeError as re:
#            if str(re).startswith("Selected KDE bandwidth is 0. Cannot estimate density."):
#                return seaborn.histplot(series[mask], kde_kws={'bw': 0.1})
#            else:
#                raise re
#        except ValueError as error:
#            if str(error).startswith("could not convert string to float"):
#                plt.figure(figsize=(0, 0))
#                return countgrid(series, sort=True)    
#            else:
#                raise error        

                
                
                

                
                
                
                
# ==============================================================================================
# analyse_freqs 
# war: top_beispiele
# ==============================================================================================

def analyse_freqs(data, cols=None, limits=[], splits=[], sort_count=True ):
    """ 
    Frequency analysis that includes a subordinate frequency analysis. 
    Provides e.g. the most important examples per case. Splits strings and lists.
    * cols:        Columns to which the analysis is to be applied. Name or list of names.
                   If one of the addressed columns is a list, it is exploded.
    * limits:      List of limits, corresponding to cols.
                   Example: [None,5] does not limit the frequencies of the first col 
                   but limits the data for the second col to the 5 most common values.
    * splits:      List of split characters, corresponding to cols.
                   Used to explode sentences into words
    * sort_count:  True =>  result shows the most common contents first
                   False => result is sorted by group
    See the jupyter notebook for examples.
    """
    
    # generelle Fehler abfangen
    if data is None:
        return 'Nothing to analyze'
    
    if not isinstance(data, pd.DataFrame):  
        data = dataframe(data, autotranspose=1)
        data.columns = ['item']
        return analyse_freqs( data, data.columns[0], limits=limits, splits=splits, sort_count=sort_count)

    if data.shape[0] == 0:
        return 'No rows'
    
    # Parameter cols 
    if type(cols) is str:
        cols = [cols] #let the command take a string or list    
        
    # Parameter limits 
    if type(limits) is int:
        limits = [limits] #let the command take an int or list  
    if len(limits) == 0:
        limits = [9999999]
    if not limits[0]:
        limits[0] = 9999999        
    if len(limits) < len(cols):
        limits += [None] * (len(cols)-len(limits)) 
    limits = [ 20 if l is None else l  for l in limits]
    #print('limits',limits)        
            
    # Parameter splits 
    if type(splits) is str:
        splits = [splits] #let the command take a string or list   
    if len(splits) < len(cols):
        splits += [None] * (len(cols)-len(splits)) 
    #print('splits',splits)
    
    # sortieren
    df = data.copy()
    #df = reset_index(data, keep_as='__afxidx').sort_values(cols + ['__afxidx']).reset_index(drop=True) 
    
    # splits realisieren
    for i, col in enumerate(cols):
        
        # Sätze zu Wortlisten
        if (splits[i] or splits[i]=='') and col:         
            df[col] = df[col].str.split(splits[i])    
    
        # Soll nach einem list-Feld gruppiert werden?
        if i == 0 and analyse_datatype(df[col])['datatype_instance'] in ['list']:
            df = df.explode(col)
    
    # Rumpf erstellen
    if len(cols) == 1:
        v = countgrid( df[cols[0]], sort=sort_count )  
        tops = v.head(limits[0])
        return tops    # nur den Rumpf zurückgeben
    else:
        v = countgrid( df[cols[0]], sort=sort_count )  #, style='top', stat=False)      
        tops = v.head(limits[0])    
    tops = drop_cols(tops,['graph'])
    
    mask        = (df[ cols[0] ].isin( tops[ cols[0] ]) )   
    feldliste   = [cols[0]] + [val                                         for val in cols[1:]   for i in (0, 1)] # erstes Feld einfach, die anderen doppelt
    funcliste   = ['group'] + [partial(top_values,       limit=lim)        if i==0 else \
                               partial(top_values_count, limit=lim)        for lim in limits[1:] for i in (0, 1)]     
    namensliste = [cols[0]] + [val                                         if i==0 else \
                               val + '_count'                              for val in cols[1:]   for i in (0, 1)]     
    
    #print(feldliste)
    #print(funcliste)
    #print(namensliste)
    
    result = group_and_agg( df[mask], feldliste, funcliste, namensliste, verbose=False )
    #return result
    return tops.merge(result)
    



                


# ==============================================================================================
# Kurze Einzelroutinen
# ==============================================================================================


#def val_first_valid(series):
#    """ 
#    Returns the first notna value of a series
#    """
#    
#    try:
#        result = series.loc[series.first_valid_index()]
#        if isinstance(result, pd.Series): # das liegt an nonunique Index
#            return result.iloc[0]
#        else:
#            return result
#    except:
#        return None
#
#
#def val_last_valid(series):
#    """
#    Returns the last notna value of a series
#    """
#    try:
#        result = series.loc[series.last_valid_index() ]
#        if isinstance(result, pd.Series): # das liegt an nonunique Index
#            return result.iloc[-1]
#        else:
#            return result        
#    except:
#        return None


    
def val_most(series):    
    """
    Returns the most frequent value of a series
    """
    mask = ~(series.isna().fillna(False))
    ohnenull = series[mask]   
    return ohnenull.value_counts().idxmax()


# wird von analyse_cols verwendet.
def nunique(series):
    """
    Returns the number of different values.
    Safe from TypeError: unhashable type: 'list'.
    """
    try:
        return series.nunique()
    except: 
        return series.apply(lambda x: str(x)).nunique()
    
    

def ntypes(series):
    '''
    Returns the number of different types. Examines all values of the series.
    NaN values are not counted.
    '''
    return series[series.notna()].map(type).nunique()





##############################################################################################
# ............................................................................................
# DataFrames 
# ............................................................................................
##############################################################################################



# sortiert die Spalten neu, vielfältigste Spalten zuerst 
def sort_cols_by_nunique(df, inaccurate_limit=100000):
    """
    Returns the DataFrame with reordered columns.
    It is sorted by nunique.
    * inaccurate_limit: If the dataframe is bigger than this, take a sample of this size
    """
    
    
    if df.shape[0] > inaccurate_limit:
        return sort_cols_by_nunique(quicksample(df, inaccurate_limit))
    
    spaltendict = dict()
    for col in df.columns:
        try:
            spaltendict[col] = df[col].nunique()
        except:
            spaltendict[col] = -1
            
    # sortieren
    spaltendict = {k: v for k, v in sorted(  spaltendict.items(), key=lambda item: item[1], reverse=True   )}

    #spalten = list(analyse_values(df, sort=True, with_index=False).col_name)
    df = df.reindex(spaltendict.keys(), axis=1)
    return df





# ==================================================================================================
# analyse_groups
# ==================================================================================================


def analyse_groups(df, exclude=[], tiefe_max=3):
    """
    Analyses a DataFrame for uniqueness and redundancy.
    Groups by many combinations of columns and counts the duplicates that are created in the process.
    Interpretation:
    0 dups => This combination of columns is unique
    Same number of dups than other combination of columns => Indication of redundancy
    
    """
    #return analyse_groups_worker(df, exclude, tiefe_max)
    try:
        return analyse_groups_worker(df, exclude, tiefe_max)
    except:
        # alles in String wandeln
        print('wandle in Strings...')
        df = df.applymap(lambda x: str(x))   
        return analyse_groups_worker(df, exclude, tiefe_max)


    
# ==================================================================================================
# same_but_different
# ==================================================================================================

def same_but_different(df, same, different, sort=True, return_mask=False):
    """ 
    Returns the rows of a DataFrame that are the same on the one hand and different on the other:
    They are the same in the fields named in same.
    And they differ in the field named in different.
    This is useful for analysing whether fields correlate 100% with each other or are independent.
    * same:       Array of column names.
    * different:  Single column name.  This column is used to search for differences.
    """
    mask = df.groupby(same)[different].transform('nunique') > 1
    if return_mask:
        return mask
    if sort:
        return df[mask].sort_values(same)
    return df[mask]    
        

    
    
    

#################################################################################################
# ...............................................................................................
# Lib-interne Hilfsfunktionen 
# ...............................................................................................
#################################################################################################



# Häufigkeitsverteilung für disjunkte Daten
# sort=True:  häufigstes zuerst
# sort=False: in der Reihenfolge der Eingabedaten
def countgrid( series, sort=True ):
    if series.shape[0] == 0:
        warnings.warn('No data')
        return 
    result = pd.DataFrame(   series.value_counts()   ).reset_index()
    countname   = series.name + '_count'
    percentname = series.name + '_percent'
    result.columns = [series.name, countname]
    result[percentname] = scale(result[countname], method='rel').round(3)*100 
    result['graph'] = result[percentname].apply( lambda x: int(x*0.5)*'#')
    if sort:
        return reset_index(result.sort_values(countname, ascending=False))
    return reset_index(result.sort_values(series.name))







# wird von analyse_groups verwendet.
# Komplettanalyse auf Eindeutigkeit und Redundanz
def analyse_groups_worker(df, exclude=[], tiefe_max=3):
    
    # Wichtigste Spalten zuerst
    df = sort_cols_by_nunique(df)
    
    # alle Teilmengen einer Menge
    # liefert list 
    from itertools import chain, combinations
    def powerlist(iterable, len_min=1, len_max=0):
        "powerlist('abcd',2, 2)  -->  [['a', 'b'], ['a', 'c'], ['a', 'd'], ['b', 'c'], ['b', 'd'], ['c', 'd']]"
        s = list(iterable)
        if len_max==0:
            len_max = len(s)
        result = chain.from_iterable(combinations(s, r) for r in range(len_min,len_max+1))
        result = [list(r) for r in result]
        return result    
    
    # spalten ermitteln
    spalten = [s for s in col_names(df, without='float') if not s in exclude]
    #print(spalten)
    
    # Leeres Ergebnis
    result_0 = []
    
    # Erst die Einerkombis, dann die Zweierkombis
    fertig = 0
    menge_eindeutiger_cols = set()
    
    for level in range(1,tiefe_max+1):
        
        if level > tiefe_max:
            break        
        
        # spaltenkombis
        spaltenkombis = powerlist(spalten,level,level)

        # Einzelne Zeile
        for sp in spaltenkombis:
            
            # sinnlose Kombination?
            p = len(set(sp).intersection(menge_eindeutiger_cols))
            if p > 0:
                continue
                
            a = sp                                      # columns
            b = level                                   # level = group_size
            c = df.duplicated(subset=sp).sum()          # dups_abs
            if c == 0 and len(sp) == 1:
                menge_eindeutiger_cols.add(sp[0]) 
            anfügen = (a, b, c)
            result_0.append( (a, b, c ) )         # alles anfügen            
        
        #print(menge_eindeutiger_cols)
        
        # Ergebnis aufbereiten
        result_1 = pd.DataFrame.from_records(result_0)
        result_1.columns = ['columns','level','dups_abs']
        
        # vorzeitigen Abbruch vormerken, sobald es spaltenkombis ohne dups gibt
        if len(menge_eindeutiger_cols) > 0  and fertig==0:
            tiefe_max = level +1 # Level noch zuende machen
            fertig = 1

    
    # Vollständiges Ergebnis endgültig aufbereiten    
    result_1 = result_1.sort_values(['dups_abs','level'])
    result_1 = reset_index(result_1)
    result_1['dups_rel'] = result_1.dups_abs / df.shape[0]     
    return result_1



    
def memory_consumption( iteration_of_objects, limit=10, use_rtype=True):
    '''
    Returns the memory consumption of Python objects.
    * iteration_of_objects: can be e.g. a DataFrame or just locals()
    * limit: Limits the output size
    * use_rtype: Use rtype instead of type?
    
    For the memory consumption of the biggest 10 local variables call:
    bpy.memory_consumption( locals() )
    '''    
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        result = bpy.memory_consumption( iteration_of_objects, limit=limit, use_rtype=use_rtype)
        result = dataframe(result, verbose=False)
        if use_rtype:
            result.columns = ['name','rtype','size']
        else:
            result.columns = ['name','type','size']
        return result






import random, string
from collections import Counter

import numpy  as np
import pandas as pd 
import bpyth  as bpy

from .config   import Config
from .pandas   import reset_index, dataframe, drop_cols, rename_col, repeat
from .analyse  import nunique, change_datatype, col_names
from .compare  import check_equal
from .scale    import scale

try:
    from perlin_noise import PerlinNoise
except:
    pass







#################################################################################
# ...............................................................................
# Excel
# ...............................................................................
#################################################################################


#
def dump_excel(df, filename, filetype='xlsx', tabcol='', index=False, changedatatype=True, check=True, verbose=None ):
    """ 
    Writes a dataframe into an xlsx file for Excel or Calc.
    * tabcol:  Should the data get split into different Excel tabs using this column?
    * check:   Should the file be read in again and the result checked for identity?
    """
    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
    
    # changedatatype
    if changedatatype:        
        df = change_datatype(df, verbose=False)   
        
    # .xlsx
    if not filename.endswith('.'+filetype):
        filename += '.'+filetype
    
    # Einzelner Tab
    if tabcol == '':
        with pd.ExcelWriter(filename) as writer: 
            df.to_excel(writer, sheet_name='Pandas', index=index, inf_rep='__unendlich__', na_rep='__NaN__' )
            
    # Viele Tabs        
    else:
        alle_tabs = df[tabcol].unique().tolist()    
        with pd.ExcelWriter(filename) as writer:  
            for counter, tab in enumerate(alle_tabs):
                mask = df[tabcol] == tab
                tabname = tab #+ '___' + str(counter)
                df[mask].drop(tabcol, axis=1).to_excel(writer, sheet_name=tabname, index=index)
                
    # check
    if check:
        df2 = load_excel(filename, tabcol=tabcol)
        assert check_equal( df, df2, reset_index=True )
              

#dataframe_to_excel = dump_excel
  

                
                
# Die Umkehrung: Liefert den Dataframe zurück.   
# Wenn tabcol angegeben, wird eine zusätzliche Spalte angefügt, die den Namen des Tab enthält.
#
def load_excel(filename, filetype='xlsx', sheet_name=None, tabcol='', changedatatype=True, verbose=None):
    ''' 
    Loads a dataframe from an xlsx file (Excel or Calc).
    * sheet_name:        Only read this sheet
    * tabcol:            Should the Excel tabs get a column in the result?
    * changedatatype:    Should the datatypes get optimized?
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')  
        
    # .xlsx
    if not filename.endswith('.'+filetype):
        filename += '.'+filetype
        
    # read
    sheets_dict = pd.read_excel(filename, sheet_name=sheet_name ) 
    
    if sheet_name is not None:
        result = sheets_dict
        if changedatatype:        
            result = change_datatype(result, verbose=False)      
        for col in col_names(result, 'str'):
            result[col] = result[col].str.strip()             
        return result
    
    result = pd.DataFrame()
    
    for name, sheet in sheets_dict.items():
        if tabcol != '':
            sheet[tabcol] = name # zusätzliche Spalte
        sheet = sheet.rename(columns=lambda x: x.split('\n')[-1])
        #result = result.append(sheet)
        result = pd.concat( [result, sheet] )

    result.reset_index(inplace=True, drop=True)  
    
    if changedatatype:        
        result = change_datatype(result, verbose=False)    
    for col in col_names(result, 'str'):
        result[col] = result[col].str.strip()    
        
    if verbose:
        print(result.shape[0], 'rows loaded')          
    
    return result

#excel_to_dataframe = load_excel





#################################################################################
# ...............................................................................
# Random data
# ...............................................................................
#################################################################################


def random_series(size, typ, **kwargs):
    ''' 
    Returns a series of random data. 
    * size
    * typ: 'int', 'float', 'string', 'name', 'choice', 'list', 'mix',
           'ascending', 'descending', 'perlin' or 'errorprone'. Or the first letter of this.
           'name' generates random first names, 'list' generates lists of random first names.
           'mix' generates mixed datatypes. 
           'ascending', 'descending' and 'perlin' generates ordered random sequences.
           'errorprone' generates sequences of NaNs, 0, 1 with similar index. Useful for testing. 

    The other arguments are passed to the appropriate functions for the type of random data.
    General arguments are:
    * name
    * p_nan: value 0..1 specifies  how many NaNs are interspersed
    * p_dup: value 0..1 determines how many Dups are included.
    
    There are extra parameters for some types of random data:
    - int:    min, max: closed interval, min and max are both possible values in the generated data
    - float:  decimals: how many decimal places
    - string: len_min, len_max: controls the length of the generated character sequence
              mix: Explicit specification of the available character set. Overwrites p_dup. 
                   Example: mix='ABCabc'
    - list:   len_min, len_max: controls the length of the generated lists.                   
    - choice: choice: List or Series of elements to choose
    - perlin: freq: List of up to frequencies, see random_perlin for more details
              op: 'add' or 'mult', how the frequencies are linked together
              sc: scaling, default is 'max_abs', this scales -1..1, see function scale
    
    Examples:    
    =========
    random_series( 10, 'int')
    random_series( 10, 'string', len_min=1, len_max=2)
    '''

    if typ in   ['int','i']: 
        myfunc = random_series_int
    elif typ in ['float','f']: 
        myfunc = random_series_float   
    elif typ in ['ascending','a']: 
        myfunc = random_series_ascending   
    elif typ in ['descending','d']: 
        myfunc = random_series_descending          
    elif typ in ['perlin','p']: 
        myfunc = random_series_perlin         
    elif typ in ['string','str','s']: 
        myfunc = random_series_string 
    elif typ in ['name','n']: 
        myfunc = random_series_name  
    elif typ in ['choice','c']: 
        myfunc = random_series_choice  
    elif typ in ['list','l']: 
        myfunc = random_series_list     
    elif typ in ['mix','m']: 
        myfunc = random_series_mix     
    elif typ in ['errorprone','e']: 
        myfunc = random_series_errorprone        
    
    p_dup = kwargs.get('p_dup',0)
    p_nan = kwargs.get('p_nan',0)    
    name  = kwargs.get('name',0)    
    assert p_dup >= 0
    assert p_dup <  1
    
    # unhashable type
    if typ in ['list','l','mix','m']:
        result = myfunc(size, **kwargs) 
        if p_nan>0:
            #print('p_nan',p_nan)
            result = result.apply(decorate, p=p_nan) # mit nan dekorieren        
        return result
    
    # Keine Dups wählbar 
    if typ in ['ascending','a','descending','d','perlin','p','errorprone','e']: 
        result = myfunc(size, **kwargs)
        if p_nan>0:
            result = result.apply(decorate, p=p_nan) # mit nan dekorieren        
        return result
    
    # Keine Dups gewünscht    
    elif p_dup == 0:
        result = myfunc(size*2, **kwargs).drop_duplicates().head(size) 
        fortschritt = nunique(result)
        for i in range(1,15):
            #result = result.append(myfunc(size*2, **kwargs)).drop_duplicates().head(size)
            result = pd.concat( [result, myfunc(size*2, **kwargs)] )
            result = result.drop_duplicates().head(size)
                
            fortschritt_neu = nunique(result)
            #print(size, i, fortschritt, fortschritt_neu)
            if (fortschritt_neu == size) or (fortschritt_neu == fortschritt):
                break # Keine Dups mehr drin, oder kein weiterer Fortschritt
            fortschritt = fortschritt_neu
        # Ende    
        #result = result.append(myfunc(size, **kwargs))         # noch ein letztes Mal auffüllen        
        result = pd.concat( [result, myfunc(size, **kwargs)] )  # noch ein letztes Mal auffüllen
        result = result.reset_index(drop=True).head(size)       # beschneiden und reindexieren
        if p_nan>0:
            #print('p_nan',p_nan)
            result = result.apply(decorate, p=p_nan) # mit nan dekorieren
            if typ in  ['int','i']: 
                result = result.astype('Int64')
        if typ in  ['string','str','s','name','n']: 
            result = result.astype('string')                    
        return result
    
    # Dups gewünscht
    else:
        basis = myfunc(size*10, **kwargs).drop_duplicates().head(  int(size*(1-p_dup)  )  )      # dupfrei
        result = random_series( size, 'choice', choice=basis, p_nan=p_nan, p_dup=0, name=basis.name)
        if p_nan>0:
            result = result.apply(decorate, p=p_nan) # mit nan dekorieren         
            if typ in  ['int','i']: 
                result = result.astype('Int64')  
        if typ in  ['string','str','s','name','n']: 
            result = result.astype('string')  
        result = result.reset_index(drop=True)
        return result
    
    # Ende wird nie erreicht
    assert False
    
    
    

def decorate(skalar, p=0.2, special=np.nan):
    """ 
    Decorates a series with specials (e.g. NaNs), is applied with apply
    e.g. result = result.apply(decorate, p=0.1)               # decorate with 10% nan  
    e.g. result = result.apply(decorate, p=0.1, special='a')  # decorate with 10% 'a'.              
    """
    if p <= 0 or random.random() > p:
        return skalar
    else:
        if random.random() > 0.5:
            try:
                return skalar + [None]
            except:
                return special
        else:
            return special
        


    


# Fertige gemischte Zufallsdaten
def people(size=100):
    """ Returns a DataFrame with random data for testing purposes """
    a = random_series( size, 'name',                                                 p_dup=0.3,  name='first_name' )
    b = random_series( size, 'int',    min=20, max=30) + random_series( size, 'int', min=0, max=12) 
    b.name = 'age'
    bb = b.apply(lambda x: int(x/10)*10 )  
    bb.name = 'age_class'    
    c = random_series( size, 'int',    min=10000, max=99999,            p_nan=0.02,  p_dup=0.3,  name='postal_code')    
    d = random_series( size, 'choice', choice=['Bremen','Berlin'],      p_nan=0.3,   p_dup=0,    name='birthplace')
    e = random_series( size, 'string', len_min=5, len_max=10,           p_nan=0,     p_dup=0,    name='secret')
    f = random_series( size, 'string', len_min=0, len_max=5,            p_nan=0,     p_dup=0.2,  name='features').apply(set)    
    g = random_series( size, 'choice', choice=['ABC','ABCC','','abc','cba','Ax','AAA','ACCB','bbab'],  name='history').apply(list)  
    result = dataframe( (a,b,bb,c,d,e,f,g), verbose=False )  
    result = change_datatype(result,        verbose=False)
    return result       



# Fertige gemischte Zufallsdaten
def random_numbers(size=1000):
    """ Returns a DataFrame with random data for testing purposes """
    a = random_series( size, 'int',    min=20, max=30) + random_series( size, 'int', min=0, max=12) 
    b = random_series( size, 'int',    min=10000, max=99999,  p_dup=0.3)
    c = random_series( size, 'float',  decimals=3, p_dup=0.3)    
    d = random_series( size, 'float',  decimals=3, p_dup=0.3) * 10   
    e = random_series( size, 'ascending')   
    result = dataframe( (a,b,c,d,e),   verbose=False ) 
    result = change_datatype(result, verbose=False)    
    return result  

zufallsdaten = random_numbers
randomdata   = random_numbers
        

    
    
def random_perlin( shape=(100,5), freq=[3,6,12,24], op='add'):
    '''
    Returns a Dataframe with Perlin Noise:
    - Every column looks like a random walk.
    - The columns correlate (the noise is 2-dim)
    * shape: The shape of the result
    * freq:  Up to 4 frequencies for the random walk.
             The frequencies are added or multipied with descending factors 1, 1/2, 1/4 and 1/16.
             A frequency of 1 means 1 maximum and 1 minimum.
             A frequency of 2 means 2 maxima  and 2 minima.                
             A frequency of 0 means, that nothing is added for the corresponding factor.
             E.g. freq=[1,0,0,100] gives a very low-frequency random walk 
             with a very slight admixture of high-frequency components. 
    * op:    'add' or 'mult', how the frequencies are linked together
                
    '''
    if isinstance(freq,int):
        freq = [freq]
    freq = list(freq) + [1, 1, 1, 1]
    freq = freq[:4]
    freq_nn = [ o if o>0  else 1 for o in freq ]
    noise1 = PerlinNoise(octaves=freq_nn[0])
    noise2 = PerlinNoise(octaves=freq_nn[1])
    noise3 = PerlinNoise(octaves=freq_nn[2])
    noise4 = PerlinNoise(octaves=freq_nn[3])

    xpix, ypix = shape
    pic = []
    for i in range(xpix):
        row = []
        for j in range(ypix):
            noise_val = noise1([i/xpix, j/ypix])
            if freq_nn[1] > 0:
                nv2 = 0.5 * noise2([i/xpix, j/ypix])
                if op=='add':
                    noise_val += nv2
                elif op=='mult':
                    noise_val *= nv2                   
            if freq_nn[2] > 0:
                nv2 = 0.25 * noise3([i/xpix, j/ypix])
                if op=='add':
                    noise_val += nv2
                elif op=='mult':
                    noise_val *= nv2  
            if freq_nn[3] > 0:                
                nv2 = 0.0625 * noise4([i/xpix, j/ypix])
                if op=='add':
                    noise_val += nv2
                elif op=='mult':
                    noise_val *= nv2                  

            row.append(noise_val)
        pic.append(row)
    return dataframe(pic, verbose=False)





# ==================================================================================================
# Interne Funktionen für zufällige Testdaten
# ==================================================================================================
    

def random_series_int( size, min=0, max=1000, name='rnd_int', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random integers between min and max. min and max are both possible.
    '''
    result = pd.Series(np.random.randint(min, max+1, size))
    result = result.rename(name)        
    return result



def random_series_float(size, decimals=3, name='rnd_float', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random floats between 0 and 1 with specified decimal places. 
    '''
    f1 = 10**decimals
    result = pd.Series(np.random.randint(0, f1, size) / f1)  
    result = result.rename(name)     
    return result



def random_series_perlin(size, name='rnd_perlin', freq=[3,6,12,24], op='add', sc='max_abs', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random floats between -1 and 1 
    '''
    result = random_perlin( shape=(size,1), freq=freq, op=op ).A
    if sc:
        result = scale( result, sc)
    result = result.rename(name)     
    return result



# Liefert eine Series zufälliger Floats zwischen 0 und 1 
def random_series_monotonic(size, name='rnd_monotonic', ascending=True, index=None, p_nan=0, p_dup=0):
    '''
    Returns an monotonic, crooked Series of random floats
    '''
    # freq=[ 0.1, 0.2, 3,0]
    f1 = random.uniform( 0.01, 0.2 )
    f2 = random.uniform( 0.05, 0.2 )  
    f3 = random.uniform( 2.0,  4.0 )  
    freq = [f1,f2,f3,0]
    result = random_series_perlin( size, freq=freq, op='mult', sc='min_max')
    result = result.sort_values(ascending=ascending).reset_index(drop=True)
    result = result.rename(name)    
    return result



def random_series_ascending(size, name='rnd_ascending', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random floats between 0 and 1 
    '''
    result = random_series_monotonic(size=size, name=name, ascending=True)
    return result



def random_series_descending(size, name='rnd_descending', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random floats between 0 and 1 
    '''
    result = random_series_monotonic(size=size, name=name, ascending=False)
    return result



# .
def random_series_string(size, len_min=4, len_max=7, name='rnd_string', index=None, p_nan=0, p_dup=0, mix=None):
    '''
    Returns a series of random strings.
    * size: Length of the series
    * len_min: Minimum length of the random strings 
    * len_max: Maximum length of the random strings  
    * name: Name of the series
    * p_nan: Dead parameter
    * p_dup: 0..1, influences the width of the character set and thus the dup probability
    * mix: Explicit specification of the available character set. Overwrites p_dup. Example: mix='ABCabc'
    '''
    if not mix:
        mix = string.ascii_letters + string.digits + 'ÄÖÜäöüaeiou'
        ziel = int((1-p_dup)*len(mix))
        mix = mix[:ziel]        
    result = pd.Series([ bpy.random_str(size_min=len_min, size_max=len_max, mix=mix) for i in range(size) ])  
    result = result.rename(name) 
    result = result.astype('string')
    return result
   
    

def random_series_list(size, len_min=2, len_max=10, name='rnd_list', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random lists.
    '''
    #länge = random.randint(len_min,len_max)
    #c = list(random_series_name(länge))
    result = pd.Series([ list(random_series_name(random.randint(len_min,len_max))) for i in range(0,size)])
    
    
   #result = pd.Series(pd.util.testing.rands_array(len_max, size))    
   #result = result.str.replace('[aeiouAEIOU]',' ', regex=True)
   #
   #def bearbeite_element( skalar ):
   #    return skalar.split() 

   #result = result.apply(bearbeite_element)      
    result = result.rename(name) 
    return result    



def random_series_choice(size, choice=[], name='rnd_choice', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random selections
    '''
    if type(choice) is pd.core.series.Series:
        choice = list(choice)
    elif choice == []: 
        choice = list('abcde')
    result = pd.Series(np.random.choice(choice, size=size))      
    result = result.rename(name)     
    return result   
 


def random_series_name(size, name='rnd_name', index=None, p_nan=0, p_dup=0):
    '''
    Returns a series of random first names
    '''
    vornamen = ['Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna','Anna', 'Maria', 'Karl', 'Hans', 'Carl', 'Michael', 'Paul', 'Jan', 'Elisabeth', 'Alexander', 'Peter', 'Andre', 'Christian', 'Johanna', 'Marie', 'Thomas', 'Andreas', 'Walter', 'Johannes', 'Max', 'Werner', 'Matthias', 'Julia', 'Katharina', 'Martin', 'Daniel', 'Klaus', 'Stephan', 'Stefan', 'Claus', 'Emma', 'Christina', 'Tobias', 'Hermann', 'Wolfgang', 'Günter', 'Jürgen', 'Helmut', 'Ursula', 'Günther', 'Wilhelm', 'Heinrich', 'Tim', 'Kurt', 'Gerhard', 'Robert', 'Oliver', 'Nicole', 'Lisa', 'Heinz', 'Florian', 'Sebastian', 'Martha', 'Otto', 'Philipp', 'Eva', 'Mark', 'Horst', 'Helga', 'Sarah', 'Sven', 'Ernst', 'Markus', 'Georg', 'Erika', 'Uwe', 'Charlotte', 'Nils', 'Karin', 'Friedrich', 'Herbert', 'Lars', 'Frieda', 'Ingrid', 'Stefanie', 'Rolf', 'Nina', 'Sabine', 'Katrin', 'Susanne', 'Monika', 'Renate', 'Dennis', 'Patrick', 'Jens', 'Mathias', 'Gisela', 'Gertrud', 'Frank', 'Manfred', 'Franz', 'Marc', 'Christoph', 'Marcel', 'Bernd', 'Anja', 'Rudolf', 'Alfred', 'Jasmin', 'Lena', 'Felix', 'Alexandra', 'Harald', 'Petra', 'Willi', 'Sandra', 'Lukas', 'Melanie', 'Annika', 'Dieter', 'Claudia', 'Elke', 'Marion', 'Laura', 'Jana', 'Fritz', 'Brigitte', 'Simon', 'Christine', 'Heike', 'Barbara', 'Andrea', 'Mike', 'Julian', 'Marko', 'Gerda', 'Fabian', 'Maik', 'Joachim', 'Birgit', 'Caroline', 'Benjamin', 'Sara', 'Hanna', 'Margarete', 'Sonja', 'Johann', 'Marco', 'Rainer', 'Jessica', 'Margarethe', 'Ute', 'Richard', 'Jörg', 'Lea', 'Kai', 'Jonas', 'Ilse', 'Dominik', 'Tom', 'Timo', 'Helene', 'Hildegard', 'Jutta', 'Vanessa', 'Luise', 'Moritz', 'Lara', 'Erna', 'Niklas', 'Christa', 'Nico', 'Klara', 'Holger', 'David', 'Meik', 'Anne', 'Kerstin', 'Franziska', 'Ralf', 'Martina', 'Torsten', 'Kevin', 'Niels', 'Hannah', 'Ingeborg', 'Gerd', 'Gabriele', 'Berndt', 'Angelika', 'Lucas', 'Jannik', 'Dirk', 'Silvia', 'Rene', 'Philip', 'Phillip', 'Louise', 'Paula', 'Erich', 'Sophie', 'Marianne', 'Clara', 'Edith', 'Alina', 'Irmgard', 'Leon', 'Miriam', 'Carolin', 'Volker', 'Karsten', 'Maximilian', 'Willy', 'Leonie', 'Isabell', 'Maike', 'Lina', 'Eric', 'Tanja', 'Hannelore', 'Niclas', 'Inge', 'Christiane', 'Erik', 'Thorsten', 'Pia', 'Henry', 'Michelle', 'Walther', 'Kay', 'Yannik', 'Yannick', 'Yannic', 'Josef', 'Frida', 'Ole', 'Daniela', 'Manuela', 'Björn', 'Isabel', 'Luisa', 'Marcus', 'Ben', 'Anke', 'Jennifer', 'Finn', 'Natalie', 'Fynn', 'Sofie', 'Sascha', 'Isabelle', 'Sylvia', 'Robin', 'Emil', 'Jakob', 'Vincent', 'Kim', 'Nele', 'Elena', 'Carla', 'Bärbel', 'Gustav', 'Christel', 'Pascal', 'Karoline', 'Niko', 'Carsten', 'Malte', 'Viktoria', 'Lilly', 'Maja', 'Erwin', 'Nadine', 'Karla', 'Lilli', 'Sophia', 'Norbert', 'Antonia', 'Ruth', 'Jacob', 'Sina', 'Luka', 'Emilie', 'Ella', 'Bettina', 'Neele', 'Lennart', 'Lennard', 'Manuel', 'Victoria', 'Louisa', 'Anton', 'Katja', 'Käthe', 'Ida', 'Waltraud', 'Angela', 'Louis', 'Luca', 'Heidi', 'Anni', 'Emily', 'Celina', 'Eileen', 'Nathalie', 'Elfriede', 'Bruno', 'Marvin', 'Sabrina', 'Mia', 'Maya', 'Hendrik', 'Amelie', 'Silke', 'Karina', 'Sofia', 'Timm', 'Marina', 'Meike', 'Regina', 'Carina', 'Arthur', 'Rita', 'Melina', 'Melissa', 'Hannes', 'Sigrid', 'Bernhard', 'Hertha', 'Jonathan', 'Herta', 'Linda', 'Marlene', 'Olaf', 'Ulrich', 'Marta', 'Torben', 'Janina', 'Diana', 'Zoe', 'Astrid', 'Nick', 'Ulrike', 'Josephine', 'Swen', 'Albert', 'Leah', 'Josefine', 'Ingo', 'Elli', 'Lieselotte', 'Larissa', 'Luis', 'Annica', 'Arne', 'Jule', 'Aileen', 'Ayleen', 'Britta', 'Stephanie', 'Yvonne', 'Kathrin', 'Noah', 'Jacqueline', 'Bastian', 'Cornelia', 'Michaela', 'Joshua', 'Heiko', 'Lothar', 'Curt', 'Svenja', 'Jannick', 'Yasmin', 'Rebecca', 'Christopher', 'Chiara', 'Beate', 'Doris', 'Toni', 'Lucy', 'Antje', 'Theresa', 'Steffen', 'Leo', 'Kirsten', 'Simone', 'Mara', 'Adrian', 'Nora', 'Leoni', 'Nicolas', 'Nikolas', 'Mario', 'Catharina', 'Till', 'Axel', 'John', 'Detlef', 'Elias', 'Konstantin', 'Ina', 'Kira', 'Julius', 'Raphael', 'Merle', 'Bianca', 'Siegfried', 'Dominic', 'Jannis', 'Pauline', 'Lasse', 'Linus', 'Reinhard', 'Henri', 'Denis', 'Rosemarie', 'Benedikt', 'Celine', 'Rudolph', 'Helena', 'Annette', 'Stella', 'Evelyn', 'Tina', 'Vivien', 'Else', 'Dagmar', 'Mohammed', 'Anneliese', 'Adolf', 'Artur', 'Ali', 'Milena', 'Oskar', 'Maurice', 'Isabella', 'Phil', 'Elly', 'Mika', 'Marius', 'Fiona', 'Hugo', 'Theo', 'Anette','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Tanja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Anja','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom','Tom']  
    result = pd.Series(np.random.choice(vornamen, size=size))    
    result = result.astype('string') 
    result = result.rename(name)      
    return result    




def random_series_mix(size, name='rnd_mix', p_nan=0, p_dup=0):
    '''
    Returns a series with various datatypes
    '''
    anz = int(size / 5) + 1
    
    b = random_series( anz, 'int',    min=-66666, max=66666,          )    
    c = random_series( anz, 'float',  decimals=4,                     )
    d = random_series( anz, 'string', len_min=2, len_max=20,          )
    e = random_series( anz, 'name',                                   )
    f = random_series( anz, 'choice', choice=['Bremen','Bremerhaven'] )
    g = random_series( anz, 'list',                                   )    
      
    #result = b.append(c).append(d).append(e).append(f).append(g)
    result = pd.concat( [b, c, d, e, f, g] )
    result = result.sample(frac=1).reset_index(drop=True).head(size)
    result = result.rename(name)    
    result.iloc[0] = {0}
    
    return result



def random_series_errorprone(size, name=None, values=None, index=None, p_nan=0, p_dup=0):
    '''
    Provides a random series of known issues
    '''
    if values is None:
        values = random.randint(0, 9)
    if index is None:
        index = random.randint(-1, 9)

    if values == 0:
        result =  repeat( [0], size )     
    elif values == 1:
        result =  repeat( [1], size )
    elif values == 2:
        result =  repeat( [-1], size )         
    elif values == 3:
        result =  repeat( [None], size )  
    elif values == 4:
        result =  repeat( [np.NaN], size )  
    elif values == 5:
        result =  repeat( [-1,1], size )   
    elif values == 6:
        result =  repeat( [0,-1], size )
    elif values == 7:
        result =  repeat( [0,1], size )  
    elif values == 8:
        result =  repeat( [0,np.NaN], size )    
    elif values == 9:
        result =  repeat( [42], size )              
    else:
        result =  repeat( [42], size ) 

    # name    
    if name is None:
        result.name = 'rnd_errorprone' + '_' + str(values)
    else:
        result.name = name

    # index
    if index == -1:
        return result 
    else:
        result.index = random_series_errorprone(size,values=index, index=-1)  
        result.index.name = None
        if result.name != name:
            result.name  += 'i' + str(index)              
        return result
            









import numpy  as np
import pandas as pd

try:
    import seaborn
except ImportError:
    pass
    #print('no seaborn, plot will not work')  
##seaborn.set()
    
import matplotlib.pyplot as plt   
import warnings, logging

from .config   import Config
from .pandas   import rename_col
from .subsets  import sample

    

    
# ==============================================================================================
# plot
# ==============================================================================================

# Todo: automatisch ausdünnen, wenn zuviele Datenpunkte
#
def plot(   df1, df2=None, x='--', size=(16, 4), palette=('rainbow','tab10'), line=(1,1), inaccurate_limit=10000   ):
    ''' 
    Plots DataFrames or Series.
    * df1, df2:         The first two parameters are DataFrames or Series. 
                        If there are two, they get separate y-axes.
    * x:                Which column contains the x-axis? 
                        x='index' -> The index is used as x.    
                        If no x is given, x is tried to be guessed. 
                        If no suitable column is found, the index is used as x.
                        A column is considered suitable if it is called 'x', 'X' or 'index'.
    * size:             Width and height of the plot as tuples
    * palette:          The two palettes as tuple or sting
    * line:             The line thickness as tuple or number    
    * inaccurate_limit: From what size should the data be thinned randomly.
                        Uses pandasklars sample function, so minimums and maximums are kept. 
    '''   
        
    logging.getLogger('matplotlib.font_manager').disabled = True
    
    if isinstance(df1, pd.Series):
        df1 = pd.DataFrame(df1).reset_index()
        x='index'
        
    if isinstance(df2, pd.Series):      
        df2 = pd.DataFrame(df2).reset_index() 
    
    # ausdünnen, falls nötig
    df1 = sample(df1,inaccurate_limit)
    df2 = sample(df2,inaccurate_limit)    
    
    # Einspaltiges DataFrame?
    #if df1.shape[1] == 1:       
    #    df1 = df1.reset_index()
    #    col0 = list(df1.columns)[0]
    #    df1 = rename_col(df1, col0, 'x')   
    #    x = 'x'
    #
    #if not df2 is None:
    #    if df2.shape[1] == 1:               
    #        df2 = df2.reset_index()
    #        col0 = list(df2.columns)[0]
    #        df2 = rename_col(df2, col0, 'x')   
    #        #x = 'x2'
    
    # Kein x festgelegt aber es gibt eine Spalte namens x, X oder index
    schnittmenge = set(df1.columns)  &  {'x','X','index'}
    if x=='--':
        schnittmenge = set(df1.columns)  &  {'x','X','index'}        
        if len(schnittmenge)==1:
            x = next(iter(schnittmenge))  # x festlegen
    
    # Erste Spalte untersuchen
    if x=='--':
        col0 = list(df1.columns)[0]
        if list(df1[col0]) == list(df1.index):
            x = col0
        else:
            x = 'index'
      
    # x=='index'
    if x=='index' and not 'index' in df1.columns:
        df1 = df1.reset_index()            
        x='index'
        if not df2 is None:
            if not 'index' in df2.columns:
                df2 = df2.reset_index()              
   
    #print('x =',x)        
        
    # palette1 und palette2  festlegen    
    if isinstance(palette, tuple):
        palette1 = palette[0]
        palette2 = palette[1]
    else:
        palette1 = palette
        
    # line1 und line2  festlegen    
    if isinstance(line, tuple):
        line1 = line[0]
        line2 = line[1]
    else:
        line1 = line      
    
    # Erste y-Achse  

    plt.figure(figsize=size)
    seaborn.lineplot(x=x, 
                 y='value', 
                 hue='variable', 
                 palette=palette1,
                 linewidth=line1,
                 data=pd.melt(df1, [x])
                )       
    plt.legend(loc='upper left')
    plt.ylabel( ' '.join([str(c) for c in df1.columns[1:]]) )
    
    if df2 is None:
        return
    
    # Zweite y-Achse
    ax2 = plt.twinx()
    seaborn.lineplot(x=x, 
                 y='value', 
                 hue='variable', 
                 palette=palette2,
                 linewidth=line2,                 
                 data=pd.melt(df2, [x]),
                 ax=ax2
                )     

    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)  
    plt.ylabel( ' '.join([str(c) for c in df2.columns[1:]]) )
    
                

        
# ==============================================================================================
# histo
# ==============================================================================================
        
def histo(series, quantile=1):                
    '''Histogramm'''
    logging.getLogger('matplotlib.font_manager').disabled = True
    mask = (series <= series.quantile(quantile))   &   (series >= series.quantile(1-quantile)) 
    try:
        plt.figure(figsize=(16, 4))
        return seaborn.histplot(series[mask])
    except RuntimeError as re:
        if str(re).startswith("Selected KDE bandwidth is 0. Cannot estimate density."):
            return seaborn.histplot(series[mask], kde_kws={'bw': 0.1})
        else:
            raise re
    except ValueError as error:
        if str(error).startswith("could not convert string to float"):
            plt.figure(figsize=(0, 0))
            return countgrid(series, sort=True)    
        else:
            raise error                        







import warnings, copy
from functools import partial

import pandas as pd 
import numpy  as np
import bpyth  as bpy
   
from collections import Counter, defaultdict 

from .config     import Config
from .pandas     import dataframe, reset_index, drop_cols, rename_col, move_cols, drop_multiindex, quicksample, add_rows, move_rows
from .type_info  import type_info


###############################################################################################
# .............................................................................................
# Aggregation ...
# .............................................................................................
###############################################################################################


def group_and_agg(df, col_origins, col_funcs=None, col_names=None, dropna=True, optimize=False, verbose=None): 
    '''
    Groups and aggregates. Provides a user interface similar to that of MS Access.
    * col_origins: list of all columns to process
    * col_funcs:   list of all functions to apply to the columns above. 
                   Sometimes you have to use strings, sometimes function names.
                   'group' or '' = grouping. 
    * col_names:   list of new names for the result columns. Optional. Space = default name will be taken.
    * dropna:      Parameter for groupby.
    * optimize:    True to handle duplicated rows seperatly. 
                   Useful in situations with not many duplicated rows and slow functions in col_funcs.
    
    Example:
    df = pak.people()
    pak.group_and_agg( df, 
                       col_origins=['age_class', 'birthplace', 'first_name',  'age', 'age', 'first_name'],
                       col_funcs  =['group',     'group',      pak.agg_words, 'min', 'max', 'min'],
                 )    
    '''
    
    if verbose is None:
        verbose = Config.get('VERBOSE')      
    
    # Steuertabelle bauen
    steuer = dataframe((col_origins,  col_funcs,  col_names), verbose=False)
    steuer.columns =  ['col_origins','col_funcs','col_names']
    #return steuer
    steuer.col_funcs = steuer.col_funcs.fillna('')
    mask = (steuer['col_funcs'].str.len() == 0)   
    steuer.loc[mask,'col_funcs'] = 'group'   
    
    # cols_group: Die Spalten nach denen gruppiert werden soll
    mask = (steuer.col_funcs == 'group')
    cols_group = list(steuer[mask].col_origins)    
    
    g = steuer.groupby('col_origins')    
    s = g.agg(list) #.reset_index()
    s = s.drop(cols_group)
    d = s.to_dict()['col_funcs'] # dict für agg
    
    # name_new errät den Namen des Feldes, das die agg-Funktion ausspuckt
    steuer['name_new'] = steuer.col_origins 
    mask = steuer.col_funcs.apply(lambda x: isinstance(x, str))   &   (steuer.col_funcs != 'group')
    steuer.loc[mask,'name_new'] += '_' + steuer[mask].col_funcs
    mask = steuer.col_funcs.apply(lambda x: callable(x))  
    steuer.loc[mask,'name_new'] += '_' + steuer[mask].col_funcs.astype('str').str.split(' ').str[1].str.replace('>','').str.replace("'",'').str.replace('Series.','',regex=False).str.replace('.','',regex=False) 
    reihenfolge = list(steuer.name_new)

    if col_names:
        mask = (steuer['col_names'].str.len() == 0)
        steuer.loc[mask,'col_names'] = steuer[mask].name_new # Standardnamen übernehmen
        
    if optimize:
        if df.isnull().values.any():
            raise ValueError('Works without NaNs only')
        mask = df.duplicated(subset=cols_group, keep=False)
        df_unique, df_withdups = move_rows(df, mask, msg=None, verbose=False)
        if verbose:
            print( '{0} unique rows and {1} rows with duplicates'.format(df_unique.shape[0] , df_withdups.shape[0])  )
        assert df_unique.shape[0] + df_withdups.shape[0] == df.shape[0]
        result = group_and_agg( df_withdups, col_origins=col_origins, col_funcs=col_funcs, col_names=col_names, dropna=dropna, optimize=False, verbose=False)
        result = add_rows(result,df_unique[col_origins],verbose=False)
        result = reset_index(result)
        if verbose:
            n0 = df.shape[0]
            n1 = result.shape[0]        
            print( '{0} rows less, now {1} rows'.format(n0-n1, n1)  )        
        return result    
    
    # result bauen
    if d:    
        gruppiert = df.groupby(cols_group, dropna=dropna)
        result = gruppiert.agg(d) 
    else:
        result = df.groupby(cols_group, dropna=dropna, as_index=False).first()[cols_group]

    result = drop_multiindex(result, verbose=False)
    if len(cols_group) == 1:
        result = result.reset_index() # in anderen Fällen macht drop_multiindex das
    #print('reihenfolge',reihenfolge)
    result = move_cols(result, reihenfolge)   
    
    if verbose:
        n0 = df.shape[0]
        n1 = result.shape[0]        
        print( '{0} rows less, now {1} rows'.format(n0-n1, n1)  )
    
    
    if not col_names:      
        return result
    #return steuer
    #print(list(result.columns))
    #print(list(steuer['col_names']))
    result.columns = list(steuer['col_names'])
    return result
    





def most_freq_elt(series, inaccurate_limit=(10000,1000) ):  
    '''
    Aggregates a Series to the most frequent scalar element.
    Like Series.mode, but always returns a scalar.
    If two elements are equally frequent, just any one is returned .
    * inaccurate_limit: If the data is bigger than this, examine take a sample of this size.
      The first value is for hashable datatypes, the second for non_hashable datatypes.
      Set inaccurate_limit=(None,None) if you don't accept inaccuraties.

    Example:
    df = pak.people()
    df.groupby('age_class')['first_name'].apply(pak.most_freq_elt)    
    ''' 
    try:
        if type_info(series).is_hashable:
            if inaccurate_limit[0] is None:
                return series.mode().iloc[0]
            if series.shape[0] > inaccurate_limit[0]:
                #print('ungenau hashable')
                return quicksample(series, inaccurate_limit[0]).mode().iloc[0]
            return series.mode().iloc[0]
            
        else:
            if inaccurate_limit[1] is None:
                return np.NaN
            if series.shape[0] > inaccurate_limit[1]:
                #print('ungenau not hashable')
                result = quicksample(series, inaccurate_limit[1])
                with warnings.catch_warnings():
                    warnings.simplefilter(action='ignore', category=UserWarning)
                    return result.mode().iloc[0]
            return series.mode().iloc[0]
    except:
        return np.NaN
        # list(series.mode())[0]
    
    

def top_values(series, limit=3, count=False):
    '''
    Aggregates a Series to a list of the most frequent elements.
    Or, if there is only one, this single element.
    Can also return the counts of the most frequent elements.    
    * limit: limits the length of the resulting list
    * count: count=False shows the most frequent elements (default)
             cont=True   shows the corresponding counts of the elements
    Example:
    df = pak.people()
    df.groupby('age_class')['first_name'].apply(pak.top_values)
    
    Caution, does not work well with very long data sets.
    There are partials preconfigured for 3, 5, 10, 20, 100, 1000 elements,
    i.e. top_values_100 or top_values_count_20
    '''
    
    liste = series.to_list()    
    try:
        c = Counter(liste)
    except TypeError:
        flat_liste = bpy.flatten(liste)
        c = Counter(flat_liste)
        
    häufig = c.most_common(limit)
    if count:
        result = [cnt for word,cnt in häufig]
    else:
        result = [word for word,cnt in häufig]
        
    # lesbarer machen    
    if len(result) == 0:
        return ''
    if len(result) == 1:
        return result[0]  
    
    return result    

def top_values_count(series, limit=3):
    return top_values(series, limit=limit, count=True)

# partials für Aufruf in Aggregationen
# Werden hier nicht als Partial definiert, weil group_and_agg die Spaltennamen dann nicht erraten kann und durcheinander kommt

def top_values_3(series):    return top_values(series, limit=3)
def top_values_5(series):    return top_values(series, limit=5)
def top_values_10(series):   return top_values(series, limit=10)
def top_values_20(series):   return top_values(series, limit=20)
def top_values_100(series):  return top_values(series, limit=100)
def top_values_1000(series): return top_values(series, limit=1000)

def top_values_count_3(series):    return top_values_count(series, limit=3)
def top_values_count_5(series):    return top_values_count(series, limit=5)
def top_values_count_10(series):   return top_values_count(series, limit=10)
def top_values_count_20(series):   return top_values_count(series, limit=20)
def top_values_count_100(series):  return top_values_count(series, limit=100)
def top_values_count_1000(series): return top_values_count(series, limit=1000)



  


def agg_words(series):
    '''
    Aggregates a Series of strings to a long string.
    A space is always placed between the elements,
    the order is preserved.
    '''
    try:
        result = ' '.join(series.fillna(''))    
        result = bpy.superstrip(result)
    except:
        result = np.NaN
    return result

# alter Name
def agg_strings(series):
    warnings.warn('agg_strings wurde umbenannt in agg_words')
    return agg_words(series)





def agg_strings_nospace(series):
    '''
    Aggregates a Series of strings into one long string.
    No separators between the substrings.
    '''
    try:
        result = ''.join(series.fillna(''))
    except:
        result = np.NaN
    return result





def agg_words_nodup(series):
    '''
    Aggregates a Series of strings (e.g. signal words) to a long string.
    A space is always placed between the individual elements,
    the order is preserved,
    duplicates are removed.
    '''
    return agg_words(  series.str.split().explode().drop_duplicates()  )


# alter Name
def agg_strings_like_set(series):
    warnings.warn('agg_strings_like_set wurde umbenannt in agg_words_nodup')
    return agg_words_nodup(series)


    


def agg_to_list(series):
    '''
    Aggregates a Series to a list. 
    Normally this can also be done with a simple 'list', but in combination with transform this does not work.
    Then agg_to_list can be used as a substitute.
    '''
    result = [series.tolist()]*len(series)
    #result = list(s.astype(int))
    return result    



def agg_dicts(series):
    '''
    Aggregates a Series of dicts to a single dict.
    If a key occurs more than once, the value is overwritten.
    '''
    result = {key: value for d in series for key, value in d.items()}
    return result



def agg_dicts_2dd(series):
    '''
    Aggregates a Series of dicts to a single defaultdict(list).
    I.e. multiple keys are allowed. The values are always lists.  
    '''    
    result = defaultdict(list)
    for d in series:
        for key, value in d.items():
            if not value in result[key]:
                result[key].append(value)  
    return result



def agg_defaultdicts(series):
    '''
    Aggregates a Series of defaultdict(list) to a single defaultdict(list).
    '''    
    result = defaultdict(list)
    for d in series:
        for key in d:
            result[key].extend( d[key] )  
            result[key] = list(dict.fromkeys( result[key] )) # Dups entfernen
            
    return result





# ==================================================================================================
# dict
# ==================================================================================================
 

    

def explode_dict(df, col_dict, col_key='key', col_value='value', from_defaultdict=False):
    '''
    Like pandas explode, but for a dict.
    Turns dictionaries into two columns (key, value) and additional rows, if needed.
    * col_dict:          name of the column that contains the dict to explode
    * col_key:           name of the new column for the keys of the dict
    * col_value:         name of the new column for the values of the dict   
    * from_defaultdict:  Should an additional explode be executed? 
                         This can be useful for defaultdicts. Otherwise you get lists.
    '''
    if not df.index.is_unique:
        raise ValueError( 'index must be unique!')    
    result = pd.DataFrame([*df[col_dict]], df.index).stack().rename_axis([None,col_key]).reset_index(1, name=col_value)
    result = df.join(result)
    result = drop_cols(result,col_dict)
    
    if from_defaultdict:
        return result.explode(col_value)
    else:
        return result  
    
    

    
 
    
def implode_to_dict(df, cols_group=None, col_key=None, col_value=None, col_result=None, use_defaultdict=False):
    '''
    Reversal of explode_dict.
    Groups rows and turns two columns (key, value) into one dict. 
    * cols_group       is a string or list of names by which to group.
                       This affects the width of the results. None=no grouping.
    * col_key          is the name of the column containing the keys.
    * col_value        is the name of the column containing the values
    * col_result       is the name of the column containing the results
    * use_defaultdict  specifies whether the result is a dict or a defaultdict. Default: False.
                       With use_defaultdict=True, keys that occur more than once do not overwrite each other,
                       the values are always a list.
    '''
    
    if not col_key or not col_value or not col_result:
        raise
    
    # cols_group
    if not cols_group:                                                        # None
        alle = list(df.columns)
        cols_group = [c for c in alle if c not in [col_key, col_value]  ]
    elif cols_group  and not isinstance(cols_group, list):                     # Einzelner String
        cols_group = [cols_group]    
    
    # group_and_agg
    params1 = cols_group + [col_key, col_value]
    params2 = ['group'] * len(cols_group) + [list, list]
    dfg = group_and_agg(df, params1, params2, params1)
    
    # worker für die zip-Arbeit
    def worker_defaultdict(zeile, col_key, col_value, col_result):
        my_dict = defaultdict(list)
        for k, v in zip(zeile[col_key], zeile[col_value]):
            if not v in my_dict[k]:
                my_dict[k].append(v)          
        zeile[col_result] = my_dict        
        return zeile       
    
    def worker_dict(zeile, col_key, col_value, col_result):
        zeile[col_result] = dict(zip(zeile[col_key], zeile[col_value]))
        return zeile    
    
    # worker anwenden
    if use_defaultdict:
        result = dfg.apply(worker_defaultdict, axis=1, col_key=col_key, col_value=col_value, col_result=col_result)
    else:
        result = dfg.apply(worker_dict,        axis=1, col_key=col_key, col_value=col_value, col_result=col_result)    
    
    # Ende
    result = drop_cols(result,[col_key, col_value] )
    
    return result    
    
    
# implode_to_defaultdict
implode_to_defaultdict = partial(implode_to_dict, use_defaultdict=True)       

   
    
    
def cols_to_dict(df, col_dict='', cols_add=[], use_defaultdict=False, drop=True):
    '''
    Moves columns into a dict or defaultdict.
    This is 
    * col_dict:         name of the target column. Can be empty, but may already contain a dict or defaultdict. 
    * cols_add:         Columns to be packed.
    * use_defaultdict:  Should a defaultdict be used as data structure? Otherwise keys can only occur once.
    * drop:             Should the packed columns be dropped (>> move) or not (>> copy)?
    '''
    
    # Listen erzwingen
    if cols_add  and not isinstance(cols_add, list):
        cols_add = [cols_add]        
    
    
    def worker_dict(zeile, col_dict, cols_add):
        
        # Start
        zr = {}
        if col_dict in zeile.index:
            if isinstance(zeile[col_dict], dict):
                zr = dict(zeile[col_dict]) # copy                    
        
        # cols dazurechnen
        for col in cols_add:
            try:
                if (zeile[col] and not pd.isna(zeile[col]))  or  (zeile[col] == 0):
                    d = {col:zeile[col]}
                    zr.update(d)    
            except: # für Listen, die auch leer sein können
                d = {col:zeile[col]}
                zr.update(d)  
            
        zeile[col_dict] = zr
        return zeile
    
    
    
    
    def worker_defaultdict(zeile, col_dict, cols_add):
        
        # Start
        zr = defaultdict(list)
        if col_dict in zeile.index:
            if isinstance(zeile[col_dict], defaultdict):
                zr = copy.deepcopy(zeile[col_dict])
            elif isinstance(zeile[col_dict], dict):
                startvalue = { k:[v] for k,v in zeile[col_dict].items()}
                zr = defaultdict(list,startvalue)

        # cols dazurechnen
        for col in cols_add:
            if not pd.isna(zeile[col]):
                if zeile[col]  or  (zeile[col] == 0):
                    zr[col].append(zeile[col])     
            
        zeile[col_dict] = zr
        return zeile        
           
    
    
    # worker anwenden
    if use_defaultdict:
        result = df.apply(worker_defaultdict, axis=1, col_dict=col_dict, cols_add=cols_add )
    else:
        result = df.apply(worker_dict,        axis=1, col_dict=col_dict, cols_add=cols_add ) 
        
    
    # Ende
    if drop:
        result = drop_cols(result,cols_add )
    
    return result


def dict_to_defaultdict(df, col=''):
    '''
    Turns the dict in the given column into a defaultdict.
    '''
    # dirty trick
    return cols_to_dict( df, col_dict=col, use_defaultdict=True)  


def cols_to_defaultdict(df, col_dict='', cols_add=[], use_defaultdict=False, drop=True):
    '''
    Moves or copys columns into a defaultdict. 
    See cols_to_dict for more information.
    '''
    return cols_to_dict( df, col_dict=col_dict, cols_add=cols_add, use_defaultdict=True, drop=drop)  











# __init__.py

#print('pandasklar init')

import locale 
locale.setlocale(locale.LC_ALL, '') 


#Basics
from .config       import *
from .pandas       import *

# Themen
from .scale        import *
from .develop      import *
from .compare      import *
from .analyse      import *
from .plot         import *
from .content      import *
from .string       import *
from .rank         import *
from .aggregate    import *
from .subsets      import *

# Klassen
from .type_info    import type_info
from .values_info  import values_info






import re

import pandas as pd 
import numpy  as np

from bisect              import bisect_right

from .config   import Config
from .pandas   import rename_col


# import locale
# locale.setlocale(locale.LC_ALL, '') 




    





#verboten = ['zumeist ','nach ','in ']
#pat = r'\b(?:{})\b'.format('|'.join(verboten))   # 
#ALTERNATIVE: pattern = re.compile(r'\b(' + r'|'.join(verboten) + r')\b\s*')
#self.node_meta = self.node_meta.replace( pat, '', regex=True )

def remove_str(series, remove_list, safemode=True):
    '''
    Removes a list of unwanted substrings from a Series of strings.
    * remove_list: list of substrings to remove
    * safemode:    Selects the algorithm.
                   safemode=True:  Each substring is removed separately
                   safemode=False: Works with one regular expression.
                                   Special characters such as asterisks must be backslashed.    
    '''
    if safemode:
        for r in remove_list:
            series = series.str.replace(r, '', regex=False)
        return series.str.strip()    
    
    else:
        remove_list = [re.escape(w) for w in remove_list]
        pat = r'(?:{})'.format('|'.join(remove_list))   # 
        return series.str.replace( pat, '', regex=True ).str.strip()    
    
  
    
    
def remove_words(series, remove_list):
    '''
    Removes a list of unwanted words from a Series of strings.
    Works by regular expression, so special characters such as asterisks must be backslashed.  
    '''
    pat = r'\b(?:{})\b'.format('|'.join(remove_list))   # 
    return series.str.replace( pat, '', regex=True ).str.strip()        
    
    
    
def replace_str(series, translationtable):
    '''
    Replaces substrings from a Series of strings according to a translationtable.
    * translationtable: Can be a dict, a list of tuples or a DataFrame with two columns.
      Example: {'President Trump':'Trump',   'HELLO':'Hello'}
      or      [('President Trump','Trump'), ('HELLO','Hello')]
    '''
    if type(translationtable).__name__ == 'DataFrame':
        translationtable = dict(zip(translationtable.iloc[:, 0], translationtable.iloc[:, 1]))
    if not isinstance(translationtable, dict):
        translationtable = dict(translationtable )
        
    # translationtable is dict now    
    for old, new in translationtable.items():
        series = series.str.replace(old, new, regex=False)
    return series.str.strip()    
    
    
    
    


# ==================================================================================================
#  Wörter trennen
#  


def count_words(series, pat=None):
    """
    Counts the number of words (separated by space) in the strings of a series.
    Returns a series with ints.
    """
    return series.str.split(pat).str.len()    




# ==================================================================================================
#  Wörter trennen
#   

def split_col(df, col, pat=None):
    """ 
    Splits the strings of a column into individual parts, e.g. into individual words.
    The results are written in columns named A, B, C, ... consecutively.    
    - df: Dataframe
    - col: Column with strings that are to be split.
    - pat: String or regular expression to split on. If not specified, split on whitespace.
    """
    zusatzspalten = df[col].str.split(pat, expand=True)
    zusatzspalten = zusatzspalten.fillna('')
    zusatzspalten.columns = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ')[0:zusatzspalten.shape[1]]
    result = pd.concat([df, zusatzspalten], axis=1, join='inner')    
    return result
    

    
# ==================================================================================================
#  Slice
#   
    

def slice_string(df, col_text, col_start, col_end, col_result):
    """ 
    Slices a column of strings based on indexes in another columns.
    * col_text:   Name of the column containing the  text.
    * col_start:  Name of the column containing the  start index OR the start index numeric.  
    * col_end:    Name of the column containing the  end index   OR the end index numeric.      
    * col_result: Name of the column to hold the result.        
    """
    
    if isinstance(col_start, int):
        startpositions = pd.Series( col_start, index=range(df.shape[0]) )
    else:
        startpositions = df[col_start].fillna(0).astype(int)

        
    if isinstance(col_end, int):
        endpositions   = pd.Series( col_end, index=range(df.shape[0]) )
    else:        
        endpositions   = df[col_end].fillna(65535).astype(int)  
    
    df[col_result] = [ text[a:z] for text, a, z in  zip(df[col_text], startpositions, endpositions)  ]
    df[col_result] = df[col_result].astype('string')
    return df    




# ==================================================================================================
#  Encode integers as strings (to be able to work with strings instead of lists)
#   

def encode_int(s):
    """ Converts a series of small integers into strings"""
    num_rep = {10:'a',11:'b',12:'c',13:'d',14:'e',15:'f',16:'g',17:'h',18:'i',19:'j',
               20:'k',21:'l',22:'m',23:'n',24:'o',25:'p',26:'q',27:'r',28:'s',29:'t',
               30:'u',31:'v',32:'w',33:'x',34:'y',35:'z'}
    s = s.astype(int)
    mask = (s >= 10) 
    s.loc[mask] = s[mask].apply(lambda x: num_rep[x])   
    return s.astype(str)



def decode_int(s):
    """ Converting strings back into small integers"""    
    num_rep = {'0':0,'1':1,'2':2,'3':3,'4':4,'5':5,'6':6,'7':7,'8':8,'9':9,
               'a':10,'b':11,'c':12,'d':13,'e':14,'f':15,'g':16,'h':17,'i':18,'j':19,
               'k':20,'l':21,'m':22,'n':23,'o':24,'p':25,'q':26,'r':27,'s':28,'t':29,
               'u':30,'v':31,'w':32,'x':33,'y':34,'z':35 }    
    return s.apply(lambda x: num_rep.get(x) )  




# =================================================================================================================
# fast_startswith
# fast_endswith
# siehe Jupyter-Notepad zum Thema Pandas/Strings
#


def fast_startswith(df, col_search, col_found, searchfor, find_longest=True, find_identical=True):
    """
    Searches string columns for matching beginnings.
    Like pandas str.startswith(), but much faster for large amounts of data,
    and it returns the matching fragment. 
    * col_search:     Name of the column to be searched
    * col_found:      Names of the column into which the result is to be written
    * searchfor:      Series or List of strings to be searched for
    * find_longest:   Should the longest substring be given as the result? Otherwise the shortest.
    * find_identical: Should it be counted as a result if a string matches completely?
    """
    
    # startswith alternative, works only if all strings in searchme have the same length. Also returns the matching fragment
    def startwiths(data, searchme, find_identical):
        prefix = searchme[bisect_right(searchme, data)-1]
        if ((data!=prefix) or find_identical ) and data.startswith(prefix): 
            return prefix    
    
    search = pd.DataFrame(searchfor)
    search.columns = ['searchstring'] 
    search['len'] = search.searchstring.str.len()
    grouped = search.groupby('len')
    lengroups = grouped.agg(list).reset_index().sort_values('len', ascending=find_longest)  
    result = df.copy()
    result[col_found] = None
 
    for index, row in lengroups.iterrows():
        result[col_found].update(result[col_search].apply(startwiths, searchme=sorted(row.searchstring), find_identical=find_identical)  )  
        
    result[col_found] = result[col_found].astype('string')    
    return result



 
def fast_endswith(df, col_search, col_found, searchfor, find_longest=True, find_identical=True):
    '''
    Searches string columns for matching endings.
    Like pandas str.endswith(), but much faster for large amounts of data,
    and it returns the matching fragment. 
    * col_search:     Name of the column to be searched
    * col_found:      Names of the column into which the result is to be written
    * searchfor:      Series or list of strings to be searched for
    * find_longest:   Should the longest substring be given as the result? Otherwise the shortest.
    * find_identical: Should it be counted as a result if a string matches completely?
    ''' 
    # umkehren
    df = df.copy()
    df['ssdgzdgd'] = df[col_search].str[::-1]
    result = fast_startswith(df, 'ssdgzdgd', col_found, searchfor.str[::-1], find_longest=find_longest, find_identical=find_identical)
    # Ergebnis auch umkehren
    result[col_found] = result[col_found].str[::-1].copy()
    result = result.drop('ssdgzdgd', axis=1)
    return result




def preprocess_strings( dirty, how=''):
    raise('preprocess_strings ist jetzt in bj_nlp')


